/*
 MacTools.c

 Created: January 5, 2007 3:35 PM EDST
 Revised: February 3, 2007 9:54 AM EDST

 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 *  This source code is offered for use in the public domain.  You may use,
 * modify or distribute it freely.  If changes have been made to the original
 * copy, please make it clear.
 *
 *  This code is distributed in the hope that it will be useful but WITHOUT
 * ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY DISCLAIMED.
 *  This includes but is not limited to warranties of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <macintosh_cseries.h>

#include <AppleEvents.h>
#include <Gestalt.h>
#include <StandardFile.h>

#include "MacTools.h"

/* ---------- constants */

#define kStandardWNETime			30 /* half a second */

/* ---------- globals */

char				temporary[512];

const UInt8			_out_of_memory[] = "\pThere is not enough memory allocated"
" for this application program to continue.  Trying increasing the"
" \322Prefered Size\323 in the Finder\325s \322Get Info\323 box.";

QDGlobals			qd;

RgnHandle			gpRgn = (RgnHandle)NULL;

boolean				gDone = false, gModal = false, gHasAppleEvents = false;

/* ---------- prototypes */

static void initialize_application(void);
static void initialize_events(void);
static void main_event_loop();

static void startup(void);
static void suspend(void);
static void resume(void);

static pascal OSErr handle_quit_event(AEDescList *event, AEDescList *reply, long refCon);
static pascal OSErr handle_open_application_event(AEDescList *event, AEDescList *reply, long refCon);
static pascal OSErr handle_open_document_event(AEDescList *event, AEDescList *reply, long refCon);
static pascal OSErr handle_print_document_event(AEDescList *event, AEDescList *reply, long refCon);

static OSErr required_appleevent_check(AppleEvent *event);

static void main_loop(void);

static void update_window(WindowRef window);
static void activate_window(WindowRef window);

/* returns whether the event should be handled by the system or not */
static boolean handle_mousedown(EventRecord *event, DialogRef dialog);
static boolean handle_keydown(EventRecord *event, DialogRef dialog);

/* ---------- code */

void main()
{
	initialize_application();

	if ((gpRgn=NewRgn())==(RgnHandle)NULL)
		ExitToShell();

	initialize_events();
	startup();
	main_event_loop();

	if (gpRgn!=(RgnHandle)NULL)
		DisposeRgn(gpRgn);
}

static void initialize_application()
{
	int				c = 8;

	do MoreMasters(); while(--c);

# if !TARGET_API_MAC_CARBON
	MaxApplZone();
	InitGraf((Ptr)&qd.thePort);
	InitFonts();
	InitWindows();
	InitMenus();
	TEInit();
	InitDialogs(0);
# endif

	InitCursor();
}

static void initialize_events()
{
	AEEventHandlerUPP	handler_upp;

	long			result;
	OSErr			err;

	if ((err=Gestalt(gestaltAppleEventsAttr,&result))!=noErr
			|| !(result&gestaltAppleEventsPresent))
		return;

	gHasAppleEvents = true;

	/* failure doesnt matter here; whatever doesnt work, wont work */
	if ((handler_upp=NewAEEventHandlerUPP(handle_open_document_event))!=(AEEventHandlerUPP)NULL)
		err = AEInstallEventHandler(kCoreEventClass,kAEOpenDocuments,handler_upp,0,false);

	if ((handler_upp=NewAEEventHandlerUPP(handle_print_document_event))!=(AEEventHandlerUPP)NULL)
		err = AEInstallEventHandler(kCoreEventClass,kAEPrintDocuments,handler_upp,0,false);

	if ((handler_upp=NewAEEventHandlerUPP(handle_quit_event))!=(AEEventHandlerUPP)NULL)
		err = AEInstallEventHandler(kCoreEventClass,kAEQuitApplication,handler_upp,0,false);

	if ((handler_upp=NewAEEventHandlerUPP(handle_open_application_event))!=(AEEventHandlerUPP)NULL)
		err = AEInstallEventHandler(kCoreEventClass,kAEOpenApplication,handler_upp,0,false);

	//kAENotifyRecording
}

static pascal OSErr handle_open_application_event(AEDescList *event, AEDescList *reply, long refCon)
{
#pragma unused(reply,refCon)
	OSErr			err;

	if ((err=required_appleevent_check(event))!=noErr)
		return err;

	return noErr;
}

static pascal OSErr handle_open_document_event(AEDescList *event, AEDescList *reply, long refCon)
{
#pragma unused(reply,refCon)
	AEDesc			list_desc = {'NULL', NULL};
	long			count, index;
	OSErr			err;

	/* get the list of aliases */
	if ((err=AEGetKeyDesc(event,keyDirectObject,typeAEList,&list_desc))!=noErr
			|| (err=required_appleevent_check(event))!=noErr)
		goto end;

	/*
		Now, get each file from the list and process it.
		Even though the event contains a list of alises, the Apple Event Manager
			will convert each alias to an FSSpec if we ask it to.
	*/
	if ((err=AECountItems(&list_desc,&count))!=noErr)
		goto end;
	for(index=1; index<=count; index++)
	{
		FSSpec			file;
		DescType		type;
		AEKeyword		key;
		long			size;

		if ((err=AEGetNthPtr(&list_desc,index,typeFSS,&key,&type,(Ptr)&file,
				sizeof(FSSpec),&size))!=noErr)
			break;

		/* request to open something... */
		open_document(&file);
	}
end:
	AEDisposeDesc(&list_desc);
	return err;
}

static pascal OSErr handle_print_document_event(AEDescList *event, AEDescList *reply, long refCon)
{
#pragma unused(event,reply,refCon)
	/* if you want printing, just add the necessary code and change the
		returned error to noErr when youre done */
	return errAEEventNotHandled;
}

static pascal OSErr handle_quit_event(AEDescList *event, AEDescList *reply, long refCon)
{
#pragma unused(reply,refCon)
	OSErr			err;

	if ((err=required_appleevent_check(event))!=noErr)
		return err;
	gDone = -1; /* signal that we're being asked to quit remotely */
	return noErr;
}

static OSErr required_appleevent_check(AppleEvent *event)
{
	DescType		typeCode;
	long			actual_size;
	OSErr			err;

	if ((err=AEGetAttributePtr(event,keyMissedKeywordAttr,typeWildCard,&typeCode,0,0,&actual_size))==errAEDescNotFound)
		return noErr;
	if (err!=noErr)
		return err;

	return errAEEventNotHandled;
}

/* ---------- process code */

#include "sbcp_core.h"
#include "sbcp_debug.h"

static byte random_sh__[] =
{
	_ins_pshi,0x3F,0x00,0x01,
	_ins_nop,0x00,0x00,0x00,
	_ins_exg,
	_ins_pshz,
	_ins_ret,
	_ins_stop,
	_ins_nea,0x00,0x00,0x04,
	_ins_dd,0x00,0x00,0x01,
	0xFF
};

static void startup()
{
	byte			*sh__ = random_sh__;
	int				i = 9;

Debugger();
	floor(1.4);
	floor(-1.4);
	ceil(1.4);
	ceil(-1.4);
	do
	{
		sbcp_disassemble(&sh__,temporary,temporary+64,temporary+128,sh__-random_sh__);
		fprintf(stdout,"%s  %-20s|%s\n", temporary,temporary+64,temporary+128);
	} while(--i);

	exit(0);
}

static void suspend()
{
}

static void resume()
{
	InitCursor();
	update_document(kAllDocuments);
}

static void main_event_loop()
{
	EventRecord		event;

	/* wait for the first high level event */
	do
	{
	} while(!WaitNextEvent(highLevelEventMask,&event,0,NULL));

	AEProcessAppleEvent(&event);

	do
	{
		if (gModal)
			GetNextEvent(everyEvent,&event);
		else
			WaitNextEvent(everyEvent,&event,kStandardWNETime,NULL);

		handle_event(&event,NULL);

		if (gDone && !close_document(kAllDocuments))
			gDone = 0; /* cant quit */
	} while(!gDone);
}

boolean handle_event(EventRecord *event, DialogRef dialog)
{
	WindowRef		window = (WindowRef)NULL;
	boolean			update = false, stdfilter_event = false;

	/*SystemTask();*/
	switch(event->what)
	{
	case mouseDown:
	case mouseUp:
		//if (!halted)
		if (dialog==(DialogRef)NULL)
			stdfilter_event = handle_mousedown(event,dialog);
		break;
	case keyDown:
	case keyUp:
	case autoKey:
		stdfilter_event = handle_keydown(event,dialog);
		break;
	case updateEvt:
		if ((window=(WindowRef)event->message)==(WindowRef)NULL
				|| window==(WindowRef)-1)
			break;
		if (dialog!=window)
		{
			BeginUpdate(window);
			update_window(window);
			EndUpdate(window);
		} else
			update = true;
		break;
	case activateEvt:
		window = (WindowRef)event->message;
		activate_window(window);
		break;
	case kHighLevelEvent:
		if (AEProcessAppleEvent(event)!=0)
			; /* so what if we cant handle it? */
		break;
    case osEvt:
		switch(((UInt32)event->message)>>24)
		{
		case suspendResumeMessage:
			if (event->message&resumeFlag)
				resume();
			else
				suspend();
			break;
		}
	}

	global_idle_proc();

	return update;
}

static boolean handle_mousedown(EventRecord *event, DialogRef dialog)
{
#pragma unused(event,dialog)
	return false;
}

static boolean handle_keydown(EventRecord *event, DialogRef dialog)
{
#pragma unused(dialog)
	if (event->what!=keyDown && event->what!=autoKey)
		return true;
	return false;
}

void update_window(WindowRef window)
{
#pragma unused(window)
}

void activate_window(WindowRef window)
{
#pragma unused(window)
}

void set_application_user_interaction(boolean self_only)
{
	gModal = self_only;
	AESetInteractionAllowed(self_only ? kAEInteractWithSelf:kAEInteractWithAll);
}

void global_idle_proc()
{
}

/* dialogs */

FSSpec *prompt_for_file(const UInt8 *criteria)
{
	static StandardFileReply	reply;

	if (criteria==(const UInt8*)NULL)
		criteria = empty;
	StandardGetFile(NULL,criteria[0]>>2,(const OSType*)&criteria[1],&reply);

	if (!reply.sfGood)
		return NULL;
	return &reply.sfFile;
}

static ModalFilterUPP		stdfilter = NULL;

static pascal boolean general_modal_filter_proc(DialogRef dialog,
	EventRecord *event,
	DialogItemIndex *item)
{
	short			result;

	if (GetWRefCon(GetDialogWindow(dialog))==FOUR_CHAR_CODE('askS')
			&& (event->modifiers&cmdKey))
	{
		switch(event->message&charCodeMask)
		{
		case 'N':	case 'n': case 'D':	case 'd': /* No/Dont */
			return 3;
		case 'Y':	case 'y': case 'S':	case 's': /* Yes/Save */
			return 1;
		}
	} else if (dialog!=(DialogRef)NULL && IsDialogEvent(event))
	{
		if (stdfilter==NULL && (result=GetStdFilterProc(&stdfilter))!=0)
			DebugStr("\pGetStdFilterProc gave bullsh**!");
		if (stdfilter==NULL)
		{
			return DialogSelect(event,&dialog,item);
		} else
			return InvokeModalFilterUPP(dialog,event,item,stdfilter);
	} else
		if (handle_event(event,dialog))
		{
			BeginUpdate(GetDialogWindow(dialog));
			DrawDialog(dialog);
			EndUpdate(GetDialogWindow(dialog));
		}

	return 0;
}

#if TARGET_API_MAC_CARBON
  static ModalFilterUPP				general_modal_filter_upp;
#elif GENERATINGCFM
  static RoutineDescriptor			general_modal_filter_desc =
    BUILD_ROUTINE_DESCRIPTOR(uppModalFilterProcInfo,general_modal_filter_proc);
# define general_modal_upp			(&general_modal_filter_desc)
#else
# define general_modal_upp			general_modal_filter_proc
#endif

ModalFilterUPP get_general_modal_filter_upp()
{
#if TARGET_API_MAC_CARBON
	if (general_modal_upp==NULL)
		general_modal_upp = NewModalFilterUPP(general_modal_filter_proc);
#endif
	if (general_modal_upp==NULL)
		halt();
	return general_modal_upp;
}

int ask_save(ConstStr255Param object_name)
{
	UInt8			str[128];
	int				result;

	// to prevent overflow in ParamText
	BlockMoveData(object_name,str,128);
	if (str[0]>=128 && object_name[128]!=0)
		str[127] = '';

	ParamText(str,empty,empty,empty);
	result = CautionAlert(_alert_ask_save,get_general_modal_filter_upp());
	ParamText(empty,empty,empty,empty);

	return (result+4)%3; /* reorder it into the proper result code */
}
