/*
 document.c

 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 *  This source code is offered for use in the public domain.  You may use,
 * modify or distribute it freely.  If changes have been made to the original
 * copy, please make it clear.
 *
 *  This code is distributed in the hope that it will be useful but WITHOUT
 * ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY DISCLAIMED.
 *  This includes but is not limited to warranties of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.

Wednesday, January 31, 2007 10:51:27 AM
*/

#include <macintosh_cseries.h>

#include "MacTools.h"

/* ---------- globals */

typedef struct document_record/*(16+)*/
{
	AliasHandle		alias;
	WindowRef		window;
	UInt32			flags;
	short			data_refnum, resource_refnum;
} document_record;

document_record		**document_list = (document_record**)NULL;
int					open_document_count = 0;

OSErr				last_document_error;

/* ---------- prototypes */

static WindowRef get_document_window(document_record *data);

/* ---------- code */

static boolean open_document(FSSpec *file)
{
	AliasHandle		alias;
	Ptr				storage = (Ptr)NULL;

	document_record		*data;

	WindowRef		window;
	OSErr			err;

	assert(open_document_count>=0);

	if (document_list==NULL)
	{
		if ((document_list=(struct document_record**)NewHandle(sizeof(**document_list)))==NULL)
			return false;
	} else
	{
		SetHandleSize((Handle)document_list,++open_document_count*sizeof(*document_list));
		if ((err=MemError())!=0)
		{
			alert_user(infoError,_strlist_errors,_errors_memory_new_document,err);
			return false;
		}
	}

	if (file==NULL)
		file = prompt_for_file(NULL);

	/* track the file by calling for an alias handle on it */
	/* failure would usually mean we cant access it anyway */
	if ((err=NewAlias(NULL,file,&alias))!=noErr)
		{last_document_error = err; goto kill;}

	data = (open_document_count++)+*document_list;
	data->alias = alias;
	data->flags = _document_is_dialog_window;
	data->data_refnum = data->resource_refnum = -1;

	if ((storage=NewPtr(sizeof(DialogRecord)))==NULL
			|| (data->window=(WindowRef)GetNewDialog(_dialog_document,storage,(WindowRef)-1))==NULL)
	{ /* **sigh**... kill it */
		if (storage!=(Ptr)NULL)
			DisposePtr(storage);
	kill:
		SetHandleSize((Handle)document_list,sizeof(**document_list)*--open_document_count);
		return false;
	}

	window = GetDialogWindow((DialogRef)data->window);

	/* set up crap */
	SetWTitle(window,file->name);

	/* finally bring it up */
	ShowWindow(window);
	return true;
}

/* returns true if the user didnt cancel */
boolean close_document(int index)
{
	Str255			name;
	int				count = 1;

	if (index==NONE || document_list==NULL || (*document_list)==NULL)
		return true; /* successful on... not doing anything */
	if (index==kAllDocuments)
		count = open_document_count, index = count-1; /* will be decrementing */
	do
	{
		document_record		*data = index+*document_list;

		if (open_document_count==0)
			break; /* trying to remove something from nothing? */

		if (data->flags&_document_dirty)
		{
			GetWTitle(get_document_window(data),name);
			if (!ask_save(name))
				return false;
		}

		/* kill it */
		data = index+*document_list;
		if (data->flags&_document_is_dialog_window)
			DisposeDialog((DialogRef)data->window);
		else
			DisposeWindow(data->window);

		/* remove it */
		if (index!=--open_document_count) /* move ceiling down (if not last) */
			BlockMoveData(index+1+*document_list,index+*document_list,sizeof(**document_list));
		SetHandleSize((Handle)document_list,sizeof(**document_list)*open_document_count);

		--index;
	} while(--count);

	return true;
}

boolean update_document(int index)
{
	int				count = 1;

	if (index==NONE || document_list==NULL || (*document_list)==NULL)
		return true; /* successful on... not doing anything */
	if (index==kAllDocuments)
		count = open_document_count, index = count-1; /* will be decrementing */
	do
	{
		document_record		*data = index+*document_list;

		if (open_document_count==0)
			break;
		/* update */
		--index;
	} while(--count);

	return true;
}

static WindowRef get_document_window(document_record *data)
{
	return (data->flags&_document_is_dialog_window) ? GetDialogWindow(data->window):data->window;
}
