/*
 sbcp.c
  stack-based core prototype execution section

 Version: 0.0.2
 Created: January 23, 2007 1:30 AM EDST
 Revised: February 3, 2007 12:21 PM EDST

 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 *  This source code is offered for use in the public domain.  You may use,
 * modify or distribute it freely.  If changes have been made to the original
 * copy, please make it clear.
 *
 *  This code is distributed in the hope that it will be useful but WITHOUT
 * ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY DISCLAIMED.
 *  This includes but is not limited to warranties of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.

*/

#include <cseries.h>

#include <fp.h>

#include "sbcp_core.h"
//#include "sbcp_lib.h"

#define INCREMENT_LINE_COUNTER			1

#define _nop							nop

/* ---------- constants */

enum
{
	_insx_cia = _ins_ci,

	/* 0bxx111000 */
	_insx_rct,	/* read cache table (does nothing in the emulator)					*/
				/* reads a 16 byte line for instance zero directly from cache		*/
	_insx_wct,	/* write cache table (does nothing in the emulator)					*/
				/* writes a 16 byte line for instance zero directly to cache 		*/
	_insx_setrg,/* set register						[index.s8] [reserved.16]		*/
				/* sets an internal register -- only 16 are defined					*/
	_insx_getrg,/* get register						[index.s8] [reserved.16]		*/
				/* sets an internal register -- only 16 are defined					*/
	_insx_setds,/* set DS							UInt24:address					*/
				/* sets the data stack register to an absolute address				*/
	_insx_setpc /* set PC							UInt24:address					*/
				/* sets the program counter to an absolute address					*/
};

/* ---------- macros */

#if (defined(env68k) || defined(envppc)) //&& defined(TARGET_OS_MAC)
# define UNALIGNED_RW_BE
#endif

/* these would be automatically inlined under typical circumstances */
static SInt32 read_uint24(void *s)
{
	return (((byte*)s)[0]<<16)|(((byte*)s)[1]<<8)|((byte*)s)[2];
}
static SInt32 read_sint24(void *s)
{
	return (SInt32)((((byte*)s)[0]<<24)|(((byte*)s)[1]<<16)|(((byte*)s)[2]<<8))>>8;
}

#ifdef UNALIGNED_RW_BE
/* note that the first does not work with NULL and the second before "unmapped" */
# define read_uint24(s)					(0x00FFFFFFL&*(UInt32*)(-1+(byte*)(s)))
# define read_sint24(s)					((SInt32)(*(s))>>8)
#endif

#define GEN_EXCEPTION(a,b)				do{t=a; _nop(); td=b; goto exception;}while(0)

#define CHECK_DP(p,u,o)					do{long int __tmpl; if(p>state->dc-u \
	||p<state->df+o) {t=_ev_frame_index,__tmpl=(byte*)p-state->base_pointer; \
	td=__tmpl; goto exception;}}while(0)

#define CHECK_DS(u,o)					do{long int __tmpl; \
	if(state->ds>state->dc-u) {t=_ev_stack_underflow; \
	__tmpl=(byte*)state->ds-state->base_pointer; _nop(); td=__tmpl; goto exception; \
	} else if(state->ds<state->df+o) {t=_ev_stack_overflow, \
		td=(byte*)state->ds-state->base_pointer; goto exception;}}while(0)

#define CHECK_LP(v,u,o)					do{if(v>state->lc-u) { \
	t=_ev_stack_underflow,td=(byte*)state->ls-state->base_pointer; goto exception; \
	} else if(state->ls<state->lf+o) {t=_ev_stack_overflow, \
		td=(byte*)state->ls-state->base_pointer; goto exception;}}while(0)

/* this does also a second routine for the case that the mapped page isn't writable */
#define IS_REAL_WADDRESS(s,r,w)			do{long int __tmpl=s-state->base_pointer; \
	if(__tmpl<0||((__tmpl>>=16)>state->cache_page_count) \
			|| (state->mr[__tmpl]&0x7F)==0) {r; break;} \
	if(!(__tmpb&0x80)) {w;}}while(0)

#define GET_REAL_ADDRESS(d,s,r)			do{SInt64 __tmpq=s; d=NULL; \
	RShiftS64(__tmpq,16); if(Int64N(__tmpq)||U64GeU32(__tmpq,state->cache_page_count) \
			||(state->mr[Int64ToL(__tmpq)]&0x7F)==0) {r; break;} \
	d=state->base_pointer+Int64ToL(__tmpq);}while(0)
/* this does also a second routine for the case that the mapped page isn't writable */
#define GET_REAL_WADDRESS(d,s,r,w)		do{SInt64 __tmpq=s; byte __tmpb; d=NULL; \
	RShiftS64(__tmpq,16); if(Int64N(__tmpq)||U64GeU32(__tmpq,state->cache_page_count) \
			||((__tmpb=state->mr[Int64ToL(__tmpq)])&0x7F)==0) {r; break;} \
	if(!(__tmpb&0x80)) {w;} d=state->base_pointer+Int64ToL(__tmpq);}while(0)

#define GET_FRAMEINDEX(i,d)				do{if((i)==0xFF800000L) \
	{CHECK_DS(1,0); _nop(); d=*ds++;} else {double *__tmpx; CHECK_LP(state->ls,0,1); \
	__tmpx=(double*)(state->base_pointer+Int64ToL(state->ls->return_ds)); \
	CHECK_DP(__tmpx,0,0); _nop(); d=*__tmpx;}}while(0)

#define SET_FRAMEINDEX(i,s)				do{if((i)==0xFF800000L) {CHECK_DS(0,1); --ds; \
	_nop(); ds[0]=s;} else {double *__tmpx; CHECK_LP(state->ls,0,1); \
	__tmpx=(double*)(state->base_pointer+Int64ToL(state->ls->return_ds)); \
	CHECK_DP(__tmpx,0,0); _nop(); *__tmpx=s;}}while(0)

/* ---------- prototypes */

static int sbcp_exception(sbcp_state *state, int type, double data, int err);

static int read_pointer(const sbcp_state*, SInt64 pointer, int delta, SInt64 *result);

static double fp_data_bitop(double d, double s, int what);

/* ---------- code */

int sbcp_execute_lines(sbcp_state *state, long *maximum_lines, long int remaining)
{
	SInt64			tq;
	double			*ds, td;
	byte			*s, *tp;
	long int		tl;
	long			substitute_count = LONG_MAX, t;
	int				err = 0;

	/* substitute for NULL or contents of zero */
	if (maximum_lines==(long*)NULL || *maximum_lines<0)
		maximum_lines = &substitute_count;

	/* check for any invalid pointers */
	if ((s=state->pc)==(byte*)NULL || state->df==(double*)NULL || state->mr==(byte*)NULL
			|| state->lf==(link_entry*)NULL || (ds=state->ds)==(double*)NULL)
		return _err_null_pointer;
	/* check for bad positions */
	if (ds<state->df || ds>state->dc || state->ls<state->lf || state->ls>state->lc)
		return _err_bad_stack_position;
	/* check for anything else */
	if (state->df>state->dc || state->lf>state->lc)
		return _err_floor_above_ceiling;
	if (*maximum_lines<=0)
		return err; /* still valid */

	do
	{
		/* update the state */
		state->pc = s; state->ds = ds;
		if (s==state->base_pointer) /* equivalent of NULL in the emulated environment */
			goto illegal_address;
		tl = s-state->base_pointer;
		if (tl<0 || ((tl>>=16)>state->cache_page_count) || (state->mr[tl]&0x7F)==0)
			goto illegal_address;

#if INCREMENT_LINE_COUNTER
		if (*s&0x40)
			*maximum_lines -= 1;
#endif
		if (*s&0x80)
		{
			if ((err=sbcp_exception(state,_ev_trap_vector,0,0))!=0)
				break;
			s = state->pc;
		}

		switch(*s++&0x3F)
		{
		case _ins_stop:
			remaining = 1;
			*maximum_lines = NONE; /* to signal that execution has finished */
			continue;
		case _ins_ret:
			CHECK_LP(state->ls,1,0);
			state->pc = state->base_pointer+Int64ToL(state->ls->return_pc);
			state->ds = (double*)(state->base_pointer+Int64ToL(state->ls->return_ds));
			++state->ls;
			continue;
		case _ins_exg:
			CHECK_DS(2,0);
			_nop(); td = ds[0]; ds[0] = ds[1]; _nop(); ds[1] = td;
			continue;
		case _ins_nop:
			t = read_sint24(s); s += 3;
#if INCREMENT_LINE_COUNTER
			if (t>0)
				*maximum_lines -= t;
#endif
		case _ins_dd:
			t = read_uint24(s); s += 3;
			ds += t;
			CHECK_DS(0,0);
			continue;
		case _ins_nea:
			CHECK_DS(0,1);
			t = read_sint24(s); s += 3;
			tl = s-state->base_pointer-4+t;
			--ds; _nop(); ds[0] = tl;
			continue;
		case _ins_pshz:
			CHECK_DS(0,1);
			--ds; _nop(); ds[0] = 0;
			continue;
		case _ins_pshi:
			CHECK_DS(0,1);
			t = read_uint24(s); s += 3;
			td = scalb(0x00010000L|(t&0x0000FFFFL),((t>>16)&0x7F)-79);
			if (t&0x00800000L)
				{_nop(); td = -td;}
			--ds; _nop(); ds[0] = td;
			continue;
		case _ins_setrs:
			t = read_sint24(s); s += 3;
			GET_FRAMEINDEX(t,state->rs);
			continue;
		case _ins_getrs:
			t = read_sint24(s); s += 3;
			SET_FRAMEINDEX(t,state->rs);
			continue;
		case _ins_pshd:
			CHECK_DS(1,1);
			t = read_sint24(s); s += 3;
			--ds;
			if (t==0xFF800000L)
				{_nop(); ds[0] = ds[1]; continue;}
			GET_FRAMEINDEX(t,ds[0]);
			continue;
		case _ins_popd:
			CHECK_DS(1,0);
			t = read_sint24(s); s += 3;
			if (t==0xFF800000L)
				{++ds; continue;}
			SET_FRAMEINDEX(t,*ds++);
			continue;
		case _ins_cmpd:
			CHECK_DS(1,0);
			GET_FRAMEINDEX(t,td);
			_nop();
			if (td>ds[0])
				{ds[0] = 1; continue;}
			_nop();
			if (td<ds[0])
				{ds[0] = -1; continue;}
			_nop();
			if (td==ds[0])
				{ds[0] = 0; continue;}
			_nop();
			if (ds[0]==0)
				ds[0] = td;
			continue;
		case _ins_addc:
			CHECK_DS(1,0);
			GET_FRAMEINDEX(t,td);
			_nop(); td += ds[0];
			goto set_pshc;
		case _ins_orc:
		case _ins_andc:
			CHECK_DS(1,0);
			GET_FRAMEINDEX(t,td);
			td = fp_data_bitop(td,ds[0],s[-1]&0x3F);
		set_pshc:
			SET_FRAMEINDEX(t,td);
			_nop();
			if (td==0)
				ds[0] = -1;
			else
				ds[0] = 0;
			continue;
		case _ins_rdd:
		case _ins_rds:
		case _ins_rdb:
		case _ins_rdfd:
		case _ins_rddu:
		case _ins_rdsu:
		case _ins_rdbu:
		case _ins_rdfe:
			t = read_uint24(s); s += 3;
			CHECK_DS(0,1);
			--ds; _nop(); ds[0] = 0;
			_nop(); td = 0;
			if (t&0x00800000L)
				break;
			tq = state->instance[(t>>16)&0x7F];
			Add64S32(tq,(signed)(t<<16)>>16);
			GET_REAL_ADDRESS(tp,tq,_nop(); td=1);
			_nop();
			if (td!=0)
				goto access_fault;
			switch(s[-1]&0x3F)
			{
			case _ins_rddu:
				state->instance[(t>>16)&0x7F] = tq;
			case _ins_rdd:
				_nop(); td = *(double*)tp;
				break;
			case _ins_rdsu:
				state->instance[(t>>16)&0x7F] = tq;
			case _ins_rds:
				_nop(); td = *(float*)tp;
				break;
			case _ins_rdbu:
				state->instance[(t>>16)&0x7F] = tq;
			case _ins_rdb:
				_nop(); td = *tp;
				break;
			case _ins_rdfd:
				break;
			case _ins_rdfe:
				break;
			}
			--ds; _nop(); ds[0] = td;
			continue;
		case _ins_wdd:
		case _ins_wds:
		case _ins_wdb:
		case _ins_wdfd:
		case _ins_wddu:
		case _ins_wdsu:
		case _ins_wdbu:
		case _ins_wdfe:
			t = read_uint24(s); s += 3;
			CHECK_DS(1,0);
			_nop(); td = 0;
			if (t&0x00800000L)
				break;
			tq = state->instance[(t>>16)&0x7F];
			Add64S32(tq,(signed)(t<<16)>>16);
			GET_REAL_WADDRESS(tp,tq,_nop(); td=1,_nop(); td=2);
			_nop();
			if (td!=0)
			{
				_nop();
				if (td==2)
				{
				write_fault:
					tl = tp-state->base_pointer;
					GEN_EXCEPTION(_ev_write_fault,tl);
				} else
					goto access_fault;
			}
			_nop(); td = *ds++;
			switch(s[-1]&0x3F)
			{
			case _ins_wddu:
				state->instance[(t>>16)&0x7F] = tq;
			case _ins_wdd:
				_nop(); *(double*)tp = td;
				break;
			case _ins_wdsu:
				state->instance[(t>>16)&0x7F] = tq;
			case _ins_wds:
				_nop(); *(float*)tp = td;
				break;
			case _ins_wdbu:
				state->instance[(t>>16)&0x7F] = tq;
			case _ins_wdb:
				_nop(); *tp = td;
				break;
			case _ins_wdfd:
				break;
			case _ins_wdfe:
				break;
			}
			continue;
		case _ins_or:
		case _ins_and:
		case _ins_eor:
			CHECK_DS(1,0);
			ds[1] = fp_data_bitop(ds[1],ds[0],s[-1]&0x3F); ++ds;
			continue;
		case _ins_sz: CHECK_DS(1,0); _nop(); ds[0] = -(ds[0]==0); continue;
		case _ins_sgt: CHECK_DS(1,0); _nop(); ds[0] = -(ds[0]>0); continue;
		case _ins_sge: CHECK_DS(1,0); _nop(); ds[0] = -(ds[0]>=0); continue;
		case _ins_slt: CHECK_DS(1,0); _nop(); ds[0] = -(ds[0]<0); continue;
		case _ins_sle: CHECK_DS(1,0); _nop(); ds[0] = -(ds[0]<=0); continue;
		case _ins_add: CHECK_DS(1,0); _nop(); ds[1] += ds[0]; ++ds; continue;
		case _ins_sub: CHECK_DS(1,0); _nop(); ds[1] -= ds[0]; ++ds; continue;
		case _ins_mul: CHECK_DS(1,0); _nop(); ds[1] *= ds[0]; ++ds; continue;
		case _ins_div:
			CHECK_DS(1,0);
			_nop();
			if (ds[0]==0)
			{
				_nop();
				if (ds[0]==0)
					td = nan("004");
				else
					td = __inf();
				_nop(); ds[1] = copysign(td,ds[0]*ds[1]); ++ds;
				GEN_EXCEPTION(_ev_divide_by_zero,ds[1]);
			}
			_nop();
			ds[1] /= ds[0]; ++ds;
			continue;
		case _ins_abs: CHECK_DS(1,0); ds[0] = fabs(ds[0]); continue;
		case _ins_scal: CHECK_DS(1,0); ds[1] = scalb(ds[1],ds[0]); ++ds; continue;
		case _ins_rndz: /* trunc isn't always reliable here; floor and ceil are used */
			CHECK_DS(1,0);
			_nop();
			if (ds[0]<0)
				ds[0] = ceil(ds[0]);
			else
				ds[0] = floor(ds[0]);
			continue;
		case _ins_sqrt: CHECK_DS(1,0); ds[0] = sqrt(ds[0]); continue;
		case _ins_jnc:
			CHECK_DS(1,0);
			_nop();
			if (*ds++==0)
				continue;
		case _ins_jn:
			t = read_sint24(s); s += 3;
			s += t-4;
			continue;
		case _ins_cnc:
			CHECK_DS(1,0);
			_nop();
			if (*ds++==0)
				continue;
		case _ins_cn:
			//CHECK_LP(state->ls,0,1);
			if (state->ls>state->lf)
			{
				link_entry	*le = --state->ls;

				tl = ((byte*)state->pc)-state->base_pointer;
				LSetInt64(le->return_pc,tl);
				tl = ((byte*)state->ds)-state->base_pointer;
				LSetInt64(le->return_ds,tl);
			} else
				GEN_EXCEPTION(_ev_stack_overflow,(byte*)state->ls-state->base_pointer);
			t = read_sint24(s); s += 3;
			s += t-4;
			continue;
		case _ins_iad:
			t = read_uint24(s); s += 3;
			CHECK_DS(0,1);
			if (t&0x00800000L)
				break;
			tq = state->instance[(t>>16)&0x7F];
			Add64S32(tq,(signed)(t<<16)>>16);
			--ds; _nop(); ds[0] = Int64ToX(tq);
			continue;
		case _ins_iar:
			t = read_uint24(s); s += 3;
			CHECK_DS(0,1);
			if (t&0x00800000L)
				break;
			_nop(); td = *ds++;
			_nop(); td += ((signed)(t<<16)>>16);
			_nop(); XSetInt64(state->instance[(t>>16)&0x7F],td);
			continue;
		case _insx_rct:
		case _insx_wct:
			t = read_uint24(s); s += 3;
			continue;
		case _insx_setrg:
		case _insx_getrg:
			t = read_uint24(s); s += 3;
			if ((t&0x00F0FFFFL)!=0)
				break;
			if ((t|=(s[-4]&1)<<15)==0)
			{
				{CHECK_DS(0,1); _nop(); td = *ds++;}
				if (!(state->sr&0x40000000L))
					GEN_EXCEPTION(_ev_privileged,td);
			} else
				{CHECK_DS(1,0); --ds; _nop(); ds[0] = 0; _nop(); td = 0;}
			if (floor(td)!=td)
				GEN_EXCEPTION(_ev_non_integeral,td);
			tp = state->base_pointer+(long int)td;
			/* for protection reasons, the floor, ceiling and map values are unchanged */
			switch(t>>15)
			{
			case 0: /* set DS */
				CHECK_DP((double*)tp,0,0);
				state->ds = (double*)tp;
				break;
			case 1: /* set LS */
				CHECK_LP((link_entry*)tp,0,0);
				state->ls = (link_entry*)tp;
				break;
			case 2: state->sr = 0xFF000000UL&(UInt32)tp; break; /* set SR */
			case 3: state->pc = (unsigned long int)tp+state->base_pointer; break; /* set PC */
			case 4: /* set DF */
			case 5: /* set LF */
			case 6: /* set DC */
			case 7: /* set LC */
			case 13: /* set V1 */
				return _err_sensitive_register;
			case 16: tl = (byte*)ds-state->base_pointer; _nop(); td = tl; break; /* get DS */
			case 17: tl = (byte*)state->ls-state->base_pointer; _nop(); td = tl; break; /* get LS */
			case 18: _nop(); td = state->sr; break; /* get SR */
			case 19: tl = s-state->base_pointer; _nop(); td = tl; break; /* get PC */
			case 20: tl = (byte*)state->df-state->base_pointer; _nop(); td = tl; break; /* get DF */
			case 21: tl = (byte*)state->lf-state->base_pointer; _nop(); td = tl; break; /* get LF */
			case 22: tl = (byte*)state->dc-state->base_pointer; _nop(); td = tl; break; /* get DC */
			case 23: tl = (byte*)state->lc-state->base_pointer; _nop(); td = tl; break; /* get LC */
			case 25: _nop(); td = state->cs; break; /* get CS */
			case 26: tl = state->mr-state->base_pointer; _nop(); td = tl; break; /* get MR */
			case 27: _nop(); td = state->vs; break; /* get VS */
#if 1
			case 29: _nop(); td = Int64ToX(state->v1); break; /* get V1 */
#else
			case 29: tl = Int64ToL(state->v1); _nop(); td = tl; break; /* get V1 */
#endif
			}

			if (t&0x00008000L)
				{_nop(); ds[0] = td;}
			continue;
		case _insx_setds:
		//case _insx_setls:
		case _insx_setpc:
			tl = read_uint24(s); s += 3;
			if (!(state->sr&0x40000000L))
				GEN_EXCEPTION(_ev_privileged,tl);
			tp = state->base_pointer+tl;
			if ((t=s[-4])==_insx_setds)
			{
				CHECK_DP((double*)tp,0,0);
				state->ds = (double*)tp;
			} else if (t==_insx_setpc)
			{
				if (tl==0)
					goto illegal_address;
				state->pc = tl+state->base_pointer;
			}
			continue;
		case _ins_ci:
			if (t&0x00800000L)
				break;
			//CHECK_LP(0,1);
			if (state->ls>state->lf)
			{
				link_entry	*le = --state->ls;

				tl = ((byte*)state->pc)-state->base_pointer;
				LSetInt64(le->return_pc,tl);
				tl = ((byte*)state->ds)-state->base_pointer;
				LSetInt64(le->return_ds,tl);
			} else
				GEN_EXCEPTION(_ev_stack_overflow,(byte*)state->ls-state->base_pointer);
			td = 0; tq = state->instance[(t>>16)&0x7F];
			Add64S32(tq,(signed)(t<<16)>>16);
			GET_REAL_ADDRESS(tp,tq,td=1);
			if (td!=0)
				goto access_fault;
			if (tp!=(byte*)NULL && tp!=state->base_pointer)
				{s = tp; continue;}
		illegal_address:
			tl = tp-state->base_pointer;
			GEN_EXCEPTION(_ev_illegal_address,tl);
			continue;
		}
		sbcp_exception(state,_ev_illegal_opcode,0,0);
		continue;
	access_fault:
		tl = tp-state->base_pointer;
		t = _ev_access_fault; _nop(); td = tl;
	exception:
		err = sbcp_exception(state,t,td,0); s = state->pc;
		if (err!=0)
			return err;
#if INCREMENT_LINE_COUNTER
	} while((*maximum_lines>0 && remaining<=0) || --remaining!=0);
#else
	} while(remaining<=0 || --remaining!=0);
#endif
	state->pc = s;

	return err;
}

/* ---------- private code */

static int sbcp_exception(sbcp_state *state, int type, double data, int err)
{
	SInt64			new_position;
	long int		tl;
	int				what = type, appending_data_count = 0;

	switch(what)
	{
	case _ev_stack_overflow:
	case _ev_stack_underflow:
		if (!(state->sr&(long)0x80000000))
		{
			/* over/underflow exceptions are not generated when bit 31 is zero
			  because the floor and ceiling values are ignored */
			return _err_bad_stack_position;
		}
		/* over/underflow exceptions don't affect DS here */
		break;
	case _ev_access_fault:
	case _ev_write_fault:
	case _ev_illegal_opcode:
	case _ev_illegal_address:
	case _ev_frame_index:
	case _ev_non_integeral:
		appending_data_count = 1;
		break;
	}

	state->sr |= 0x40000000UL;
	if (state->ls>state->lf)
	{
		link_entry	*le = --state->ls;

		tl = ((byte*)state->pc)-state->base_pointer;
		LSetInt64(le->return_pc,tl);
		tl = ((byte*)state->ds)-state->base_pointer;
		LSetInt64(le->return_ds,tl);
	} else /* sucks if we can't link */
		what = _ev_execution_fault;

	if (what==_ev_execution_fault || state->ds<=state->df+appending_data_count)
	{
		//state->sr &= 0x7FFFFFFFL;
		what = _ev_execution_fault;
	} else
		*--(state->ds) = what;

	if ((err=read_pointer(state,state->v1,what*8,&new_position))!=0)
	{ /* something really, really bad just happened */

		/* execute internal flash-ROM code? */
		return err;
	}

	GET_REAL_ADDRESS(state->pc,new_position,break);
	if (state->pc==(byte*)NULL || state->pc==state->base_pointer)
		if (err==0) /* can't have more than two pending exceptions */
			err = sbcp_exception(state,what,data,type);

	/* write the information regarding the exception */
	switch(what)
	{
	case _ev_access_fault:
	case _ev_write_fault:
	case _ev_illegal_opcode:
	case _ev_illegal_address:
	case _ev_frame_index:
	case _ev_non_integeral:
		*--(state->ds) = data;
	}

	return err;
}

static int read_pointer(const sbcp_state *state,
	SInt64 pointer,
	int delta,
	SInt64 *result)
{
	byte			*real_address;

	/* first, test to see if the data in emulated space is actually mapped */
	Add64S32(pointer,delta);
	GET_REAL_ADDRESS(real_address,pointer,return _ev_access_fault);

	if (result!=NULL)
		*result = *(SInt64*)real_address;
	return 0;
}

static double fp_data_bitop(double d, double s, int what)
{
#if 0
	UInt64			bd = *(UInt64*)&d, bs = *(UInt64*)&s;
	int				xpd = (Int64Hi(bd)>>20)&0x07FF;
	int				xps = (Int64Hi(bs)>>20)&0x07FF;
	boolean			gd = Int64N(bd), gs = Int64N(bs);

	if (xpd==0x07FF)
		return d; /* destination is NAN/INF */
	if (xps==0x07FF)
		return s;  /* source is NAN/INF */

	/* expose their mantissas for what they really are */
	Int64And(bd,0x000FFFFFL,-1L);
	Int64And(bs,0x000FFFFFL,-1L);
	Int64Or(bd,0x00100000L,0L);
	Int64Or(bs,0x00100000L,0L);

	/* raise 'em to the top */
	LShiftU64(bd,11);
	LShiftU64(bs,11);

	/* make `d' the larger; adjust `s' to match bit placement */
	if (xps>xpd)
		{SWAP(bd,bs); SWAP(xpd,xps); SWAP(gd,gs);}
	if ((xpd-=xps)>127)
		xpd = 127;
	RShiftU64(bs,xpd);

	if (gd)
		UNeg64(bd);
	if (gs)
		UNeg64(bs);

	switch(what)
	{
	case _ins_or: BitOr64(bd,bs); break;
	case _ins_and: BitAnd64(bd,bs); break;
	case _ins_eor: BitExOr64(bd,bs); break;
	}

	//if (Int64N(bd))
	//	{UNeg64(bd); bd = !bd;}

	xpd += xps;
	*d = scalb(Int64ToX(bd),xpd-63-0x03FF);
#endif
	SInt64			bd, bs;

	_nop(); XSetInt64(bd,d);
	_nop(); XSetInt64(bs,s);
	if (((Int64Hi(bd)>>20)&0x07FF)==0x07FF)
		return d; /* destination is NAN/INF */
	if (((Int64Hi(bs)>>20)&0x07FF)==0x07FF)
		return s;  /* source is NAN/INF */

	switch(what)
	{
	case _ins_or: BitOr64(bd,bs); break;
	case _ins_and: BitAnd64(bd,bs); break;
	case _ins_eor: BitExOr64(bd,bs); break;
	}

	_nop();
	return Int64ToX(bd);
}
