/*
 my_io.c

 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 *  This source code is offered for use in the public domain.  You may use,
 * modify or distribute it freely.
 *
 *  This code is distributed in the hope that it will be useful but WITHOUT
 * ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY DISCLAIMED.
 *  This includes but is not limited to warranties of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.
*/

#include <cseries.h>
#include <cs_math.h>
#include <cs_strings.h>
#include <cs_system.h>

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include <Resources.h>

/* constants */

#if !defined(PROGRAM_SIGNATURE)
# define PROGRAM_SIGNATURE				"MacTools"
#endif

static const char	_log_date[] = " - %Y*%m*%d %H:%M:%S GMT\015";

/* -------- globals */

char			temporary[512]; /* 512 bytes to avoid overflow */

#ifdef PROGRAM_SIGNATURE
static const char	opening_str[] = "\015Opening log for " PROGRAM_SIGNATURE;
#else
static const char	opening_str[] = "\015Opening log for %s";
#endif

static const char	closing_str[] = "\015Closing log";

boolean			sys_debugger_available = FALSE, use_sys_debugger = FALSE;
boolean			halt_termination = FALSE; /* set by _debugMsg(...,-1) */

/* -------- debugging code */

#include <Folders.h>
#include <MacMemory.h>

/* their fault, not mine */
 /* to prevent "dprintf" from being defined as hex inline */
#undef SIXWORDINLINE
#define SIXWORDINLINE(a,b,c,d,e,f)

#include <Debugging.h>

#ifdef env68k
# pragma parameter __D0 Get68KInterruptLevel()
  extern short Get68KInterruptLevel(void) = {0x40C0,0xE048,0x240,7};
# undef ATTASKLEVEL0
# define ATTASKLEVEL0()					(Get68KInterruptLevel()==0)
#elif envppc
# ifndef ATTASKLEVEL0
#  define ATTASKLEVEL0()				(TaskLevel()==0)
# endif
#else
# define ATTASKLEVEL0()					TRUE		/* ?? */
#endif

#undef exit
 void exit(int);

static UInt8		_dbgTemp[512];		/* debug messaging temporary */

static FILE			*debugging_log = (FILE*)NULL;

/* passing FALSE means all non-urgent debugging messages (asserts, warns and dprintfs)
	go to a file named "debugging_log"; not calling this means total silence
*/
void initialize_debugger(boolean sysOnly)
{
#ifndef PROGRAM_SIGNATURE
  extern const char *_program_signature;
#endif
	close_debugging_log(); /* just in case */

	sys_debugger_available = ISLOWLEVELDEBUGGERCALLABLE() ? TRUE:FALSE;

	if (sysOnly)
	{
		Handle		h;

		if ((h=GetResource('dbug',128))==(Handle)NULL)
			use_sys_debugger = FALSE;
		else
		{
			use_sys_debugger = (GetHandleSize(h)>1) ? sys_debugger_available:FALSE;
			ReleaseResource(h);
		}
	} else if (debugging_log==(FILE*)NULL)
	{
		static boolean	first_time = TRUE;

		time_t		t;

		/* if the file doesn't exist, make it exist */
		if ((debugging_log=fopen("debugging_log","r+"))==(FILE*)NULL)
			if ((debugging_log=fopen("debugging_log","wa"))==(FILE*)NULL)
				return;

		if (first_time)
		{
			atexit(close_debugging_log);
			first_time = FALSE;
		}

		/* append the "opening" string and the date */
		fseek(debugging_log,0,SEEK_END);
		ftell(debugging_log);
#ifdef PROGRAM_SIGNATURE
		fputs(opening_str,debugging_log);
#else
		fprintf(debugging_log,opening_str, _program_signature);
#endif
		time(&t);
		strftime((char*)_dbgTemp,100,_log_date,gmtime(&t));
		fputs((char*)_dbgTemp,debugging_log);
	}
}

void close_debugging_log()
{
	time_t			t;

	if (debugging_log==(FILE*)NULL)
		return;

	/* flush the pipe so we don't render before it */
	//fflush(debugging_log);

	fputs(closing_str,debugging_log);
	time(&t);
	strftime((char*)_dbgTemp,100,_log_date,gmtime(&t));
	fputs((char*)_dbgTemp,debugging_log);
	fclose(debugging_log);
	debugging_log = NULL;
}

void debugger_log(const void *s, int z)
{
	char			buf[102];
	time_t			t;

	if (debugging_log==(FILE*)NULL)
		return;

	time(&t);
	strftime(&buf[1],100,_log_date,gmtime(&t));
	buf[0] = '\015';
	fputs(buf,debugging_log);
	fwrite(s,1,z,debugging_log);
}

/*
 [v]warn():
	to signal a warning if the expression doesn't qualify (and continue)
 [v]assert():
	typically in commercial programs, this would force a "halt", but programs used in
	project WILL NOT, so non-experienced users should terminate and (also recommended
	that they restart the computer in case of data corruption in the system)
 [v]halt():
	forced termination (probably due to data corruption where this is necessary)
*/
void _debugMsg(const char *str, const char *file, int line, int level)
{
	const char		*what = level>=0 ? (level==0 ? "warning":"err"):"halt";

	if (str==(const char*)NULL || str==(const char*)-1 || strcmp(str,"FALSE")==0)
		str = (const char*)empty;

	psprintf(_dbgTemp,str[0]==0 ?
		"file \"%s\", %s at line #%d":"file \"%s\", %s at line #%d: %s",file,what,line,str);

	if (sys_debugger_available) // && DEBUG
	{
		if (level==0) /* not pausing; add ';g' to go immediately after printing */
			*(UInt16*)&_dbgTemp[(_dbgTemp[0]=MIN(_dbgTemp[0],253)+2)-2] = ';g';
		DebugStr(_dbgTemp);
	} else
		debugger_log(&_dbgTemp[1],_dbgTemp[0]);

	if (level<0)
	{
		halt_termination = TRUE;
		exit(1);
		DebugStr("\pFailed to exit.  You must manually terminate.");
	}
}

/* ------- io stuff */

#ifdef THINK_C
void exit(int x)
{
#pragma unused(x)
	ExitToShell();
}

/* make sure that this prototype is defined */
extern int cs_exit(int);

#endif /*defined(THINK_C)*/

/* ------- printf */

void dprintf(const char *form, ...)
{
	va_list			args;
	char			*s;
	int				len;

	va_start(args,form);
#if VSNPRINTF_SUPPORTED
	vsnprintf((char*)&_dbgTemp[1],256,form,args);
	if ((len=_dbgTemp[1])>255)
		len = 255;
#else
	if ((len=vsprintf((char*)&_dbgTemp[1],form,args))>255)
		len = 255;
#endif
	va_end(args);

	s = memchr(&_dbgTemp[1],';',_dbgTemp[0]=len);

	/* if first ';' follows "g;" */
	if (debugging_log==NULL || !ATTASKLEVEL0() || s==NULL || s[1]!='g' || s[2]!=';')
	{
		if (sys_debugger_available)
		{
			DebugStr(_dbgTemp);
		} else
		{
			/* we are screwed */
		}
	} else
		debugger_log(&_dbgTemp[1],len);
}

UInt8 *psprintf(void *d, const char *form, ...)
{
	va_list			args;
	int				len;

	va_start(args,form);
	if ((len=vsprintf((char*)(1+(byte*)d),form,args))>255)
		len = 255;
	va_end(args);
	*(char*)d = len;

	return((UInt8*)d);
}

char *csprintf(char *d, const char *form, ...)
{
	va_list			args;

	va_start(args,form);
	vsprintf(d,form,args);
	va_end(args);

	return(d);
}
