/*
 sbcp.c
  stack-based core prototype execution section

 Version: 0.0.4
 Created: January 23, 2007 1:30 AM EDST
 Revised: February 5, 2007 12:57 PM EDST

 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 *  This source code is offered for use in the public domain.  You may use,
 * modify or distribute it freely.
 *
 *  This code is distributed in the hope that it will be useful but WITHOUT
 * ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY DISCLAIMED.
 *  This includes but is not limited to warranties of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.

*/

#include <cseries.h>

#include <fp.h>

#include "sbcp_core.h"
//#include "sbcp_lib.h"

#define INCREMENT_LINE_COUNTER			1

#define _nop							nop

/* ---------- constants */

enum
{
	_insx_cia = _ins_ci,

	/* 0bxx111000 */
	_insx_rct,	/* read cache table (does nothing in the emulator)					*/
				/* reads a 16 byte line for instance zero directly from cache		*/
	_insx_wct,	/* write cache table (does nothing in the emulator)					*/
				/* writes a 16 byte line for instance zero directly to cache 		*/
	_insx_setrg,/* set register						[index.s8] [reserved.16]		*/
				/* sets an internal register -- only 16 are defined					*/
	_insx_getrg,/* get register						[index.s8] [reserved.16]		*/
				/* sets an internal register -- only 16 are defined					*/
	_insx_setds,/* set DS							UInt24:address					*/
				/* sets the data stack register to an absolute address				*/
	_insx_setpc /* set PC							UInt24:address					*/
				/* sets the program counter to an absolute address					*/
};

/* ---------- macros */

#if (defined(env68k) || defined(envppc)) //&& defined(TARGET_OS_MAC)
# define UNALIGNED_RW_BE
#endif

/* these would be automatically inlined under typical circumstances */
static UInt32 read_uint24(byte *s)
{
	return (s[0]<<16)|(s[1]<<8)|s[2];
}
static SInt32 read_sint24(byte *s)
{
	return (SInt32)((s[0]<<24)|(s[1]<<16)|(s[2]<<8))>>8;
}
static UInt32 read_int32(byte *s)
{
	return (s[0]<<24)|(s[1]<<16)|(s[2]<<8)|s[3];
}

static UInt32 read_ufield(UInt8 *s, int offset, int width)
{
	if ((width&=31)==0)
		width = 32;
	s += offset>>3; offset &= 3;

	return ((((s[0]<<24)|(s[1]<<16)|(s[2]<<8)|s[3])<<offset)|(s[4]>>(7-offset)))>>(32-width);
}
static SInt32 read_sfield(UInt8 *s, int offset, int width)
{
	if ((width&=31)==0)
		width = 32;
	s += offset>>3; offset &= 3;

	return (SInt32)((((s[0]<<24)|(s[1]<<16)|(s[2]<<8)|s[3])<<offset)|(s[4]>>(7-offset)))>>(32-width);
}
static void write_field(byte *s, int offset, int width, UInt32 value)
{
	UInt32			mask;

	if ((width&=31)==0)
		width = 32;
	s += offset>>3; offset &= 3;

	/* because `offset' may nudge the pointer into the fourth byte, we can't do it
	  all at once in 32 bits; the maximum width across byte boundaries is 39 bits */
	value <<= 32-width; /* rise it to the top */
	*s++ = (*s&~((((-128)>>width)&255)>>offset))|(value>>24+offset);

	/* the rest is done only if `width' is large enough */
	if ((width-=8)<=0)
		{width = 0; return;}

	value <<= 8; value >>= offset;
	mask = (UInt32)((SInt32)0x80000000L>>width)>>offset;
	*s++ = (*s&~(mask>>24))|(value>>24);
	*s++ = (*s&~((mask>>16)&255))|((value>>16)&255);
	*s++ = (*s&~((mask>>8)&255))|((value>>8)&255);
	*s = (*s&~(mask&255))|(value&255); /* maximum width here is seven bits */
}

#ifdef UNALIGNED_RW_BE
/* note that the first does not work with NULL and the second before "unmapped" */
# define read_uint24(s)					(0x00FFFFFFL&*(UInt32*)(-1+(byte*)(s)))
# define read_sint24(s)					((SInt32)(*(s))>>8)
# define read_uint32(s)					(0+*(UInt32*)(s))
# define read_sint32(s)					(0+*(SInt32*)(s))

# if 0//def env68k
 /* this shiznit is already built-in! */
#  pragma parameter __D2 _read_ufield(__A0, __D0, __D1)
static UInt32 _read_ufield(UInt8 *s, int offset, int width) =
	{0xE9D0, 0x2821};

#  pragma parameter __D2 _read_sfield(__A0, __D0, __D1)
static SInt32 _read_sfield(UInt8 *s, int offset, int width) =
	{0xEBD0, 0x2821};

#  pragma parameter _write_field(__A0, __D0, __D1, __D2)
static void _write_field(byte *s, int offset, int width, UInt32 value) =
	{0xEFD0, 0x2821};

#  define read_ufield(d,o,w,v)			_read_ufield(d,o,w,v)
#  define read_sfield(d,o,w,v)			_read_sfield(d,o,w,v)
#  define write_field(d,o,w,v)			_write_field(d,o,w,v)
# endif
#else
# define read_uint32(s)					read_int32(s)
# define read_sint32(s)					((SInt32)read_int32(s))
#endif

#define write_dfield(d,o,w,v,f)			do{UInt32 __tmpv; if(f) {_nop(); \
	__tmpv=(SInt32)(v);} else {_nop(); __tmpv=v;} write_field(d,o,w,__tmpv);}while(0)

#define CHECK_DP(p,u,o)					do{if(p>state->dc-u||p<state->df+o) \
	{t=(p>state->dc-u) ? _ev_stack_underflow:_ev_stack_overflow; \
	tl=(byte*)p-state->base_pointer; goto lexception;}}while(0)

#define CHECK_DS(u,o)					CHECK_DP(state->ds,u,o)

#define CHECK_LP(v,u,o)					do{if(v>state->lc-u) { \
	t=_ev_stack_underflow,tl=(byte*)state->ls-state->base_pointer; goto lexception; \
	} else if(state->ls<state->lf+o) {t=_ev_stack_overflow, \
		tl=(byte*)state->ls-state->base_pointer; goto lexception;}}while(0)

/* this does also a second routine for the case that the mapped page isn't writable */
#define IS_REAL_WADDRESS(s,r,w)			do{long int __tmpl=s-state->base_pointer; \
	if(__tmpl<0||((__tmpl>>=16)>state->cache_page_count) \
			|| (state->mr[__tmpl]&127)==0) {r; break;} \
	if(!(__tmpb&128)) {w;}}while(0)

#define GET_REAL_ADDRESS(d,s,r)			do{SInt64 __tmpq=s; d=NULL; \
	RShiftS64(__tmpq,16); if(Int64N(__tmpq)||U64GeU32(__tmpq,state->cache_page_count) \
		||(state->mr[Int64ToL(__tmpq)]&127)==0) {r; break;} \
	d=state->base_pointer+Int64ToL(__tmpq);}while(0)
/* this does also a second routine for the case that the mapped page isn't writable */
#define GET_REAL_WADDRESS(d,s,r,w)		do{SInt64 __tmpq=s; byte __tmpb; d=NULL; \
	RShiftS64(__tmpq,16); if(Int64N(__tmpq)||U64GeU32(__tmpq,state->cache_page_count) \
		||((__tmpb=state->mr[Int64ToL(__tmpq)])&127)==0) {r; break;} \
	if(!(__tmpb&128)) {w;} d=state->base_pointer+Int64ToL(__tmpq);}while(0)

#define GET_FRAMEINDEX(i,d)				do{if((i)==(long)0xFF800000) \
	{CHECK_DS(1,0); _nop(); d=dp[0]; ++state->ds;} else {double *__tmpx; \
	CHECK_LP(state->ls,0,1); __tmpx=(i)+(double*)(state->base_pointer+ \
	Int64ToL(state->ls->return_ds)); CHECK_DP(__tmpx,0,0); d=*__tmpx;}}while(0)

#define SET_FRAMEINDEX(i,s)				do{if((i)==(long)0xFF800000) {CHECK_DS(0,1); \
	dp=--state->ds; _nop(); dp[0]=s;} else {double *__tmpx; CHECK_LP(state->ls,0,1); \
	__tmpx=(i)+(double*)(state->base_pointer+Int64ToL(state->ls->return_ds)); \
	CHECK_DP(__tmpx,0,0); _nop(); *__tmpx=s;}}while(0)

/* ---------- prototypes */

static int read_instance_data(sbcp_state *state, UInt32 opcode, int x, SInt64 pointer);
static int write_instance_data(sbcp_state *state, UInt32 opcode, int x, SInt64 pointer);

static int getset_register(sbcp_state *state, int w, double v);

static int do_exception(sbcp_state *state, int type, double data, int err);

static int read_pointer(const sbcp_state*, SInt64 pointer, int delta, SInt64 *result);

static double fp_data_bitop(double d, double s, int what);

/* ---------- code */

int sbcp_execute_lines(sbcp_state *state, long *maximum_lines, long int remaining)
{
	SInt64			tw;
	double			*dp, td;
	byte			*s, *tp;
	long int		tl;
	long			substitute_count = LONG_MAX, t = 0;
	int				err = 0;

	/* substitute for NULL or contents of zero */
	if (maximum_lines==(long*)NULL || *maximum_lines<0)
		maximum_lines = &substitute_count;
	/* check for any invalid pointers */
	if ((s=state->pc)==(byte*)NULL || state->df==(double*)NULL || state->mr==(byte*)NULL
			|| state->lf==(link_entry*)NULL || state->ds==(double*)NULL)
		return _err_null_pointer;
	/* check for bad positions */
	if (state->ds<state->df || state->ds>state->dc || state->ls<state->lf
			|| state->ls>state->lc)
		return _err_bad_stack_position;
	/* check for anything else */
	if (state->df>state->dc || state->lf>state->lc)
		return _err_floor_above_ceiling;
	if (*maximum_lines<=0)
		return err; /* still valid */
	_nop(); td = 0;
	do
	{
		/* update the state */
		state->pc = s; dp = state->ds;
		if (s==state->base_pointer) /* equivalent of NULL in the emulated environment */
			{tp = s; goto illegal_address;}
		tl = s-state->base_pointer;
		if (tl<0 || ((tl>>=16)>state->cache_page_count) || (state->mr[tl]&127)==0)
			{tp = s; goto illegal_address;}

#if INCREMENT_LINE_COUNTER
		if (*s&0x40)
			*maximum_lines -= 1;
#endif
		if (*s&0x80)
		{
			if ((err=do_exception(state,_ev_trap_vector,0,0))!=0)
				break;
			s = state->pc;
		}

		switch(*s++&0x3F)
		{
		case _ins_nop:
			t = read_sint24(s); s += 3;
#if INCREMENT_LINE_COUNTER
			if (t>0)
				*maximum_lines -= t;
#endif
			continue;
		case _ins_ads:
			t = read_uint24(s); s += 3; dp += t;
			CHECK_DP(dp,0,0);
			CHECK_LP(state->ls,0,1);
			if ((tl=(byte*)dp-state->base_pointer)>=Int64ToL(state->ls->return_ds))
				{t = _ev_outside_frame; goto lexception;}
			state->ds = dp;
			continue;
		case _ins_pshi:
			t = read_uint24(s); s += 3;
			CHECK_DS(0,1); --state->ds;
			td = scalb(0x00010000L|(t&0x0000FFFFL),((t>>16)&127)-79);
			if (t&0x00800000L)
				{_nop(); td = -td;}
			_nop(); dp[-1] = td;
			continue;
		case _ins_nea:
			t = read_sint24(s); s += 3;
			CHECK_DS(0,1); --state->ds;
			tl = s-state->base_pointer-4+t; _nop(); dp[-1] = tl;
			continue;
		case _ins_stop:
			remaining = 1;
			*maximum_lines = NONE; /* to signal that execution has finished */
			continue;
		case _ins_ret:
			CHECK_LP(state->ls,1,0);
			state->pc = state->base_pointer+Int64ToL(state->ls->return_pc);
			state->ds = (double*)(state->base_pointer+Int64ToL(state->ls->return_ds));
			++state->ls;
			continue;
		case _ins_pshz: CHECK_DS(0,1); --state->ds; _nop(); dp[-1] = 0; continue;
		case _ins_exg:
			CHECK_DS(2,0);
			_nop(); td = dp[0]; dp[0] = dp[1]; _nop(); dp[1] = td;
			continue;
		case _ins_setrs:
			t = read_sint24(s); s += 3;
			GET_FRAMEINDEX(t,state->rs);
			continue;
		case _ins_getrs:
			t = read_sint24(s); s += 3;
			SET_FRAMEINDEX(t,state->rs);
			continue;
		case _ins_pshd:
			t = read_sint24(s); s += 3;
			CHECK_DS(1,1); --state->ds;
			if (t==(long)0xFF800000)
				{_nop(); dp[-1] = dp[0]; continue;}
			GET_FRAMEINDEX(t,dp[0]);
			continue;
		case _ins_popd:
			t = read_sint24(s); s += 3;
			CHECK_DS(1,0); ++state->ds;
			if (t==(long)0xFF800000)
				{++state->ds; continue;}
			SET_FRAMEINDEX(t,dp[0]);
			continue;
		case _ins_cmpd:
			CHECK_DS(1,0);
			GET_FRAMEINDEX(t,td);
			_nop(); if (td>dp[0])
				{dp[0] = 1; continue;}
			_nop(); if (td<dp[0])
				{dp[0] = -1; continue;}
			_nop(); if (td==dp[0])
				{dp[0] = 0; continue;}
			/* assuming NAN */
			_nop(); if (dp[0]==0) /* otherwise (probable) NAN in dp[0] is unchanged */
				dp[0] = td;
			continue;
		case _ins_addc:
			CHECK_DS(1,0);
			GET_FRAMEINDEX(t,td);
			_nop(); td += dp[0];
			goto set_pshc;
		case _ins_orc:
		case _ins_andc:
			CHECK_DS(1,0);
			GET_FRAMEINDEX(t,td);
			td = fp_data_bitop(td,dp[0],s[-1]&0x3F);
		set_pshc:
			SET_FRAMEINDEX(t,td);
			_nop();
			if (td==0)
				dp[0] = -1;
			else
				dp[0] = 0;
			continue;
		case _ins_rdfd:
			t = read_uint32(s-1); s += 3;
			CHECK_DS(2,1); state->ds += 2;
			if (1)
			{
				SInt64		dw;

				tw = state->instance[(t>>16)&127];
				_nop(); XSetInt64(dw,dp[1]); tl = Int64Lo(dw)&7; RShiftS64(dw,3);
				SAdd64(tw,dw);
			} else
			{
		case _ins_rdd:
		case _ins_rds:
		case _ins_rdb:
		case _ins_rddu:
		case _ins_rdsu:
		case _ins_rdbu:
				t = read_uint32(s-1); s += 3;
				if (t&0x00800000L)
					break;
				CHECK_DS(0,1);
				tw = state->instance[(t>>16)&127];
			}
			if ((err=read_instance_data(state,t,tl,tw))!=0)
				return err;
			continue;
		case _ins_rdfe:
			t = read_uint32(s-1); s += 3;
			if (t&224)
				break;
			CHECK_DS(0,1); --state->ds; _nop(); dp[-1] = 0;
			tw = state->instance[(t>>16)&127]; Add64S32(tw,(signed)(t<<16)>>27);
			GET_REAL_ADDRESS(tp,tw,_nop(); tl=1);
			if (tl!=0)
				{t = _ev_access_fault, tl = tp-state->base_pointer; goto lexception;}
			if (t&0x00800000L)
				t = read_sfield(tp,(t>>8)&3,t&31);
			else
				t = read_ufield(tp,(t>>8)&3,t&31);
			_nop(); dp[-1] = t;
			continue;
		case _ins_wdfd:
			t = read_uint32(s-1); s += 3;
			CHECK_DS(3,0); state->ds += 2;
			if (1)
			{
				SInt64		dw;

				tw = state->instance[(t>>16)&127];
				_nop(); XSetInt64(dw,dp[1]); tl = Int64Lo(dw)&7; RShiftS64(dw,3);
				SAdd64(tw,dw);
			} else
			{
		case _ins_wdd:
		case _ins_wds:
		case _ins_wdb:
		case _ins_wddu:
		case _ins_wdsu:
		case _ins_wdbu:
				t = read_uint32(s-1); s += 3;
				if (t&0x00800000L)
					break;
				CHECK_DS(1,0);
				tw = state->instance[(t>>16)&127];
			}
			if ((err=write_instance_data(state,t,tl,tw))!=0)
				return err;
			continue;
		case _ins_wdfe:
			t = read_uint32(s-1); s += 3;
			if (t&224)
				break;
			CHECK_DS(1,0);
			tw = state->instance[(t>>16)&127]; Add64S32(tw,(signed)(t<<16)>>27);
			GET_REAL_WADDRESS(tp,tw,_nop(); tl=1,_nop(); tl=2);
			if (tl!=0)
				if (tl==2)
					{t = _ev_write_fault, tl = tp-state->base_pointer; goto lexception;}
				else
					{t = _ev_access_fault, tl = tp-state->base_pointer; goto lexception;}
			_nop(); td = dp[0]; ++state->ds;
			write_dfield(tp,(t>>8)&3,t&31,td,t&0x00800000L);
			continue;
		case _ins_or:
		case _ins_and:
		case _ins_eor:
			CHECK_DS(1,0);
			dp[1] = fp_data_bitop(dp[1],dp[0],s[-1]&0x3F); ++state->ds;
			continue;
		case _ins_sz: CHECK_DS(1,0); _nop(); dp[0] = -(dp[0]==0); continue;
		case _ins_sgt: CHECK_DS(1,0); _nop(); dp[0] = -(dp[0]>0); continue;
		case _ins_sge: CHECK_DS(1,0); _nop(); dp[0] = -(dp[0]>=0); continue;
		case _ins_slt: CHECK_DS(1,0); _nop(); dp[0] = -(dp[0]<0); continue;
		case _ins_sle: CHECK_DS(1,0); _nop(); dp[0] = -(dp[0]<=0); continue;
		case _ins_add: CHECK_DS(1,0); _nop(); dp[1] += dp[0]; ++state->ds; continue;
		case _ins_sub: CHECK_DS(1,0); _nop(); dp[1] -= dp[0]; ++state->ds; continue;
		case _ins_mul: CHECK_DS(1,0); _nop(); dp[1] *= dp[0]; ++state->ds; continue;
		case _ins_div:
			CHECK_DS(1,0); ++state->ds;
			_nop(); if (dp[0]==0)
			{
				_nop(); if (dp[1]==0)
					td = nan("004");
				else
					td = __inf();
				_nop(); dp[1] = copysign(td,dp[0]*dp[1]);
				t = _ev_divide_by_zero; _nop(); td = dp[1]; goto exception;
			}
			_nop(); dp[1] /= dp[0];
			continue;
		case _ins_abs: CHECK_DS(1,0); dp[0] = fabs(dp[0]); continue;
		case _ins_scal: CHECK_DS(1,0); dp[1] = scalb(dp[1],dp[0]); ++state->ds; continue;
		case _ins_rndz: /* trunc isn't always reliable here; floor and ceil are used */
			CHECK_DS(1,0);
			_nop(); if (dp[0]<0)
				dp[0] = ceil(dp[0]);
			else
				dp[0] = floor(dp[0]);
			continue;
		case _ins_sqrt: CHECK_DS(1,0); dp[0] = sqrt(dp[0]); continue;
		case _ins_jnc:
			CHECK_DS(1,0);
			_nop(); if (*(state->ds)++==0)
				continue;
		case _ins_jn:
			t = read_sint24(s); s += 3;
			s += t-4;
			continue;
		case _ins_cnc:
			CHECK_DS(1,0);
			_nop(); if (*(state->ds)++==0)
				continue;
		case _ins_cn:
			t = read_sint24(s); s += 3;
			if (state->ls>state->lf)
			{
				link_entry	*le = --state->ls;

				tl = ((byte*)state->pc)-state->base_pointer; LSetInt64(le->return_pc,tl);
				tl = ((byte*)state->ds)-state->base_pointer; LSetInt64(le->return_ds,tl);
			} else
			{
				t = _ev_stack_overflow,tl = (byte*)state->ls-state->base_pointer;
				goto lexception;
			}
			s += t-4;
			continue;
		case _ins_iad:
			t = read_uint24(s); s += 3;
			if (t&0x00800000L)
				break;
			CHECK_DS(0,1); --state->ds;
			tw = state->instance[(t>>16)&127]; Add64S32(tw,(signed)(t<<16)>>16);
			_nop(); dp[-1] = Int64ToX(tw);
			continue;
		case _ins_iar:
			t = read_uint24(s); s += 3;
			if (t&0x00800000L)
				break;
			CHECK_DS(0,1); ++state->ds;
			_nop(); XSetInt64(tw,dp[0]); Add64S32(tw,(signed)(t<<16)>>16);
			state->instance[(t>>16)&127] = tw;
			continue;
		case _insx_rct:
		case _insx_wct:
			t = read_uint24(s); s += 3;
			continue;
		case _insx_setrg:
		case _insx_getrg:
			t = read_uint24(s); s += 3;
			if ((t&0x00F0FFFFL)!=0)
				break;
			if ((err=getset_register(state,t|=(s[-4]&1)<<15,td))!=0)
				return err;
			continue;
		case _insx_setds:
		case _insx_setpc:
			tl = read_uint24(s); s += 3; tp = state->base_pointer+tl;
			if (!(state->sr&0x40000000L))
				{t = _ev_privileged; goto lexception;}
			if ((t=s[-4])==_insx_setds)
				{CHECK_DP((double*)tp,0,0); state->ds = (double*)tp;}
			else if (t==_insx_setpc)
			{
				if (tl==0)
					goto illegal_address;
				state->pc = tp;
			}
			continue;
		case _ins_ci:
			if (t&0x00800000L)
				break;
			if (state->ls>state->lf)
			{
				link_entry	*le = --state->ls;

				tl = ((byte*)state->pc)-state->base_pointer; LSetInt64(le->return_pc,tl);
				tl = ((byte*)state->ds)-state->base_pointer; LSetInt64(le->return_ds,tl);
			} else
			{
				t = _ev_stack_overflow, tl = (byte*)state->ls-state->base_pointer;
				goto lexception;
			}
			td = 0; tw = state->instance[(t>>16)&127];
			Add64S32(tw,(signed)(t<<16)>>16);
			GET_REAL_ADDRESS(tp,tw,td=1);
			if (td!=0)
				goto access_fault;
			if (tp!=(byte*)NULL && tp!=state->base_pointer)
				{s = tp; continue;}
		illegal_address:
			tl = tp-state->base_pointer;
			{t = _ev_illegal_address; goto lexception;}
			continue;
		}
		do_exception(state,_ev_illegal_opcode,0,0);
		continue;
	access_fault:
		tl = tp-state->base_pointer;
		t = _ev_access_fault;
	lexception:
		_nop(); td = tl;
	exception:
		err = do_exception(state,t,td,0); s = state->pc;
		if (err!=0)
			return err;
#if INCREMENT_LINE_COUNTER
	} while((*maximum_lines>0 && remaining<=0) || --remaining!=0);
#else
	} while(remaining<=0 || --remaining!=0);
#endif
	state->pc = s;

	return err;
}

/* ---------- private code */

static int read_instance_data(sbcp_state *state, UInt32 opcode, int t, SInt64 pointer)
{
	double			td;
	long int		tl = t;
	byte			*tp;
	int				err = 0;

	--state->ds; _nop(); *state->ds = 0; _nop(); td = 0;
	Add64S32(pointer,(signed)(opcode<<16)>>16);
	t = 0; GET_REAL_ADDRESS(tp,pointer,t=1);
	if (t!=0)
		{tl = tp-state->base_pointer; t = _ev_access_fault; goto lexception;}
	switch(((opcode>>24)&0x3F)-_ins_rdd)
	{
	case _ins_rddu-_ins_rdd:
		state->instance[(opcode>>16)&127] = pointer;
	case _ins_rdd-_ins_rdd:
		_nop(); td = *(double*)tp;
		break;
	case _ins_rdsu-_ins_rdd:
		state->instance[(opcode>>16)&127] = pointer;
	case _ins_rds-_ins_rdd:
		_nop(); td = *(float*)tp;
		break;
	case _ins_rdbu-_ins_rdd:
		state->instance[(opcode>>16)&127] = pointer;
	case _ins_rdb-_ins_rdd:
		_nop(); td = *tp;
		break;
	case _ins_rdfd-_ins_rdd:
		{
			double		dd;

			_nop(); dd = state->ds[-1]; _nop(); t = dd;
		}
		if (t<=0)
			break; /* non-positive width; result is always zero */
		if (t>=32) /* field-decode instructions clip to [0,32] */
			t = 0;
		if (opcode&0x00800000L)
			td = read_sfield(tp,tl,t);
		else
			td = read_ufield(tp,tl,t);
		break;
	lexception:
		_nop(); td = tl;
	exception:
		return do_exception(state,t,td,0);
	}
	*state->ds = td;

	return err;
}

static int write_instance_data(sbcp_state *state, UInt32 opcode, int t, SInt64 pointer)
{
	double			td;
	long int		tl = t;
	byte			*tp;
	int				err = 0;

	Add64S32(pointer,(signed)(opcode<<16)>>16);
	t = 0; GET_REAL_WADDRESS(tp,pointer,t=1,t=2);
	if (t!=0)
		if (t==2)
			{tl = tp-state->base_pointer; t = _ev_write_fault; goto lexception;}
		else
			{tl = tp-state->base_pointer; t = _ev_access_fault; goto lexception;}
	_nop(); td = *state->ds; ++state->ds;
	switch(((opcode>>24)&0x3F)-_ins_wdd)
	{
	case _ins_wddu-_ins_wdd:
		state->instance[(opcode>>16)&127] = pointer;
	case _ins_wdd-_ins_wdd:
		_nop(); *(double*)tp = td;
		break;
	case _ins_wdsu-_ins_wdd:
		state->instance[(opcode>>16)&127] = pointer;
	case _ins_wds-_ins_wdd:
		_nop(); *(float*)tp = td;
		break;
	case _ins_wdbu-_ins_wdd:
		state->instance[(opcode>>16)&127] = pointer;
	case _ins_wdb-_ins_wdd:
		_nop(); *tp = td;
		break;
	case _ins_wdfd-_ins_wdd:
		{
			double		dd;

			_nop(); dd = state->ds[-3]; _nop(); t = dd;
		}
		if (t<=0)
			break;
		if (t>=32) /* field-decode instructions clip to [0,32] */
			t = 0;
		write_dfield(tp,tl,t,td,opcode&0x00800000L);
		break;
	lexception:
		_nop(); td = tl;
	exception:
		err = do_exception(state,t,td,0);
	}

	return err;
}

static int getset_register(sbcp_state *state, int w, double v)
{
	double			td;
	byte			*tp;
	long int		tl;
	int				err = 0, t;

	if (w&0x00008000L)
	{
		{CHECK_DS(0,1); _nop(); td = *state->ds; ++state->ds;}
		if (!(state->sr&0x40000000L))
			{t=_ev_privileged; goto exception;}
	} else
		{CHECK_DS(1,0); --state->ds; _nop(); *state->ds = 0; _nop(); td = 0;}

	td = floor(v);
	_nop();
	if (v!=td)
		{t=_ev_non_integeral; goto exception;}
	tp = state->base_pointer+(long int)td;
	/* for protection reasons, the floor, ceiling and map values are unchanged */
	switch(w>>15)
	{
	case 0: /* set DS */
		CHECK_DP((double*)tp,0,0);
		state->ds = (double*)tp;
		break;
	case 1: /* set LS */
		CHECK_LP((link_entry*)tp,0,0);
		state->ls = (link_entry*)tp;
		break;
	case 2: state->sr = 0xFF000000UL&(UInt32)tp; break; /* set SR */
	case 3: state->pc = (unsigned long int)tp+state->base_pointer; break; /* set PC */
	case 4: /* set DF */
	case 5: /* set LF */
	case 6: /* set DC */
	case 7: /* set LC */
	case 13: /* set V1 */
		return _err_sensitive_register;
	case 16: tl = (byte*)(state->ds+1)-state->base_pointer; _nop(); td = tl; break; /* get DS */
	case 17: tl = (byte*)state->ls-state->base_pointer; _nop(); td = tl; break; /* get LS */
	case 18: _nop(); td = state->sr; break; /* get SR */
	case 19: tl = state->pc-state->base_pointer; _nop(); td = tl; break; /* get PC */
	case 20: tl = (byte*)state->df-state->base_pointer; _nop(); td = tl; break; /* get DF */
	case 21: tl = (byte*)state->lf-state->base_pointer; _nop(); td = tl; break; /* get LF */
	case 22: tl = (byte*)state->dc-state->base_pointer; _nop(); td = tl; break; /* get DC */
	case 23: tl = (byte*)state->lc-state->base_pointer; _nop(); td = tl; break; /* get LC */
	case 25: _nop(); td = state->cs; break; /* get CS */
	case 26: tl = state->mr-state->base_pointer; _nop(); td = tl; break; /* get MR */
	case 27: _nop(); td = state->vs; break; /* get VS */
#if 1
	case 29: _nop(); td = Int64ToX(state->v1); break; /* get V1 */
#else
	case 29: tl = Int64ToL(state->v1); _nop(); td = tl; break; /* get V1 */
#endif
	lexception:
		_nop(); td = tl;
	exception:
		err = do_exception(state,t,td,0);
	}

	if (w&0x00008000L)
		{_nop(); *state->ds = td;}

	return err;
}

static int do_exception(sbcp_state *state, int type, double data, int err)
{
	SInt64			new_position;
	long int		tl;
	int				what = type, appending_data_count = 0;

	switch(what)
	{
	case _ev_stack_overflow:
	case _ev_stack_underflow:
		if (!(state->sr&(long)0x80000000))
		{
			/* over/underflow exceptions are not generated when bit 31 is zero
			  because the floor and ceiling values are ignored */
			return _err_bad_stack_position;
		}
		/* over/underflow exceptions don't affect DS here */
		break;
	case _ev_access_fault:
	case _ev_write_fault:
	case _ev_illegal_opcode:
	case _ev_illegal_address:
	case _ev_outside_frame:
	case _ev_non_integeral:
		appending_data_count = 1;
		break;
	}

	state->sr |= 0x40000000UL;
	if (state->ls>state->lf)
	{
		link_entry	*le = --state->ls;

		tl = ((byte*)state->pc)-state->base_pointer; LSetInt64(le->return_pc,tl);
		tl = ((byte*)state->ds)-state->base_pointer; LSetInt64(le->return_ds,tl);
	} else /* sucks if we can't link */
		what = _ev_execution_fault;

	if (what==_ev_execution_fault || state->ds<=state->df+appending_data_count)
	{
		//state->sr &= 0x7FFFFFFFL;
		what = _ev_execution_fault;
	} else
		*--(state->ds) = what;

	if ((err=read_pointer(state,state->v1,what*8,&new_position))!=0)
	{ /* something really, really bad just happened */

		/* execute internal flash-ROM code? */
		return err;
	}

	GET_REAL_ADDRESS(state->pc,new_position,break);
	if (state->pc==(byte*)NULL || state->pc==state->base_pointer)
		if (err==0) /* can't have more than two pending exceptions */
			err = do_exception(state,what,data,type);

	/* write the information regarding the exception */
	switch(what)
	{
	case _ev_access_fault:
	case _ev_write_fault:
	case _ev_illegal_opcode:
	case _ev_illegal_address:
	case _ev_outside_frame:
	case _ev_non_integeral:
		*--(state->ds) = data;
	}

	return err;
}

static int read_pointer(const sbcp_state *state,
	SInt64 pointer,
	int delta,
	SInt64 *result)
{
	byte			*real_address;

	/* first, test to see if the data in emulated space is actually mapped */
	Add64S32(pointer,delta);
	GET_REAL_ADDRESS(real_address,pointer,return _ev_access_fault);

	if (result!=NULL)
		*result = *(SInt64*)real_address;
	return 0;
}

static double fp_data_bitop(double d, double s, int what)
{
#if 0
	UInt64			bd = *(UInt64*)&d, bs = *(UInt64*)&s;
	int				xpd = (Int64Hi(bd)>>20)&0x07FF;
	int				xps = (Int64Hi(bs)>>20)&0x07FF;
	boolean			gd = Int64N(bd), gs = Int64N(bs);

	if (xpd==0x07FF)
		return d; /* destination is NAN/INF */
	if (xps==0x07FF)
		return s;  /* source is NAN/INF */

	/* expose their mantissas for what they really are */
	Int64And(bd,0x000FFFFFL,-1L);
	Int64And(bs,0x000FFFFFL,-1L);
	Int64Or(bd,0x00100000L,0L);
	Int64Or(bs,0x00100000L,0L);

	/* raise 'em to the top */
	LShiftU64(bd,11);
	LShiftU64(bs,11);

	/* make `d' the larger; adjust `s' to match bit placement */
	if (xps>xpd)
		{SWAP(bd,bs); SWAP(xpd,xps); SWAP(gd,gs);}
	if ((xpd-=xps)>127)
		xpd = 127;
	RShiftU64(bs,xpd);

	if (gd)
		UNeg64(bd);
	if (gs)
		UNeg64(bs);

	switch(what)
	{
	case _ins_or: BitOr64(bd,bs); break;
	case _ins_and: BitAnd64(bd,bs); break;
	case _ins_eor: BitExOr64(bd,bs); break;
	}

	//if (Int64N(bd))
	//	{UNeg64(bd); bd = !bd;}

	xpd += xps;
	*d = scalb(Int64ToX(bd),xpd-63-0x03FF);
#endif
	SInt64			bd, bs;

	_nop(); XSetInt64(bd,d);
	_nop(); XSetInt64(bs,s);
	if (((Int64Hi(bd)>>20)&0x07FF)==0x07FF)
		return d; /* destination is NAN/INF */
	if (((Int64Hi(bs)>>20)&0x07FF)==0x07FF)
		return s;  /* source is NAN/INF */

	switch(what)
	{
	case _ins_or: BitOr64(bd,bs); break;
	case _ins_and: BitAnd64(bd,bs); break;
	case _ins_eor: BitExOr64(bd,bs); break;
	}

	_nop();
	return Int64ToX(bd);
}
