/*
 sbcp_core.h
  stack based core prototype core include

 Version: 0.0.4
 Created: January 19, 2007 9:46 PM EDST
 Revised: February 5, 2007 11:52 AM EDST

 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 *  This source code is offered for use in the public domain.  You may use,
 * modify or distribute it freely.  If changes have been made to the original
 * copy, please make it clear.
 *
 *  This code is distributed in the hope that it will be useful but WITHOUT
 * ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY DISCLAIMED.
 *  This includes but is not limited to warranties of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.

Monday, January 22, 2007 9:20:42 PM
A.P.: hmmm.  this seems pretty tishty.
*/

enum
{ /* exception vectors */
	_ev_access_fault,	/* address cannot be accessed for read/write */
	_ev_write_fault,	/* write to read-only space */
	_ev_illegal_opcode,	/* opcode is bad or unimplemented */
	_ev_illegal_address,/* bad destination */
	_ev_stack_overflow,	/* went beyond stated boundary */
	_ev_stack_underflow,/* went above ceiling */
	_ev_outside_frame,	/* stack frame index/offset is out of bounds */
	_ev_non_integeral,	/* value has a fraction; operation calls for an integer */

	_ev_trap_vector,	/* an exception vector for debugger traps */
	_ev_execution_fault,/* last resort (worst case scenario) */
	_ev_privileged,		/* operation is priviledged (can't just muck around) */
	_ev_divide_by_zero	/* division by zero */
};

enum
{ /* opcodes */
/*
  Special Effective Address types:
    UInt24              3-byte (unsigned) integer value; big-endian.

    SInt24              3-byte (signed) integer value; big-endian.

    Float24             [sign.1] [exponent.7]-63, 1. [mantissa.16]
                        NOTE: neither zero nor NAN/INF is represented.

    Field0              [sign.1] [instance.7] [offset.s16]

    Field16             [sign.1] [instance.7] [offset.s8] [reserved.3] [width.5]
                        NOTE: a width of zero is interpreted as 32.

  Instances are used for direct reads/writes.  There are 128 instances.

  The stack frame top is stored in the link frame.

  The "result" is it's own register.

  "no-ops" with a negative value may represent encoded labels.

  For stack frames, index SINT24MIN represents the current, lowest entry;
    positive indices represent parameters to the caller, and negatives
    point within the frame, itself.

  The instruction set is semi-form; either 8 or 32 bits in size.

  The first byte for instructions consists of:
    [break.1]           The instruction is set to debugger break.

    [line.1]            The instruction starts on a new source line.

    [opnumber.6]        One of the values described in the enum below.

  Any number of mnemonic operations can be achieved:
    remove one          pops 0
    repeat one          pshs 0
    boolean OR          add (only for positive values)
    boolean AND         mul
    bit NOT             pshi -1, eor
    negation            pshz, exg, sub
*/

   /* 0bxx000000    Name/Description				 Operands					*/
    _ins_nop,	/* no-operation						SInt24:linecount			*/
				/*	primarily used for source info; linecounts<0 seen as zero	*/

    _ins_ads,	/* adjust data stack				UInt24:count				*/
				/*	adds count to the data stack position						*/

    _ins_pshi,	/* push immediate					Float24:value				*/
				/*	value is pushed as a double onto the data stack				*/

    _ins_nea,	/* near effective address			SInt24:offset				*/
				/*	calculates an absolute address to the offset; pushes value	*/

    _ins_stop,	/* stop								(none)						*/
				/*	full instruction stop										*/

    _ins_ret,	/* return							(none)						*/
				/*	pulls link, resets stack frame in link with result			*/

    _ins_pshz,	/* push zero						(none)						*/
				/*	zero is pushed onto the data stack							*/

    _ins_exg,	/* stack entry exchange				(none)						*/
				/*	first two entries on data stack are exchanged				*/
   /* 0bxx001000 */
    _ins_setrs,	/* set result						SInt24:frameindex			*/
				/*	sets the result register to a stack frame index				*/

    _ins_getrs,	/* get result						SInt24:frameindex			*/
				/*	stack frame index is set to the current result				*/

    _ins_pshd,	/* push stack entry					SInt24:frameindex			*/
				/*	stack entry at top of stack frame -> data stack				*/

    _ins_popd,	/* pop to stack entry				SInt24:frameindex			*/
				/*	pulls data -> stack entry at top of stack frame				*/

    _ins_orc,	/* bitwise OR and condition			SInt24:frameindex			*/
				/*	pulls data, bit-ORs entry; pushes -1 for 0, otherwise 0		*/

    _ins_andc,	/* bitwise AND and condition		SInt24:frameindex			*/
				/*	bit-ANDs with pulled data; pushes -1 for 0, otherwise 0		*/

    _ins_cmpd,	/* compare and direction			SInt24:frameindex			*/
				/*	pulls data, compares, pushes direction: -1, 0, 1, or NAN	*/

    _ins_addc,	/* add and condition				SInt24:frameindex			*/
				/*	pulls data, adds to entry; pushes -1 for 0, otherwise 0		*/
   /* 0bxx010000 */
    _ins_rdd,	/* read data-double					[instance.s8], [offset.s16]	*/
				/*	reads a double from an instance								*/

    _ins_rds,	/* read data-single					[instance.s8], [offset.s16]	*/
				/*	reads a 32 bit "single" from an instance					*/

    _ins_rdb,	/* read data-byte					[instance.s8], [offset.s16]	*/
				/*	reads a byte from an instance								*/

    _ins_rdfd,	/* read data-field decoded			Field0						*/
				/*	reads integer by a user-defined field (pulled offset/width)	*/

    _ins_wdd,	/* write data-double				[instance.s8], [offset.s16]	*/
				/*	writes a 64 bit "double" to an instance						*/

    _ins_wds,	/* write data-single				[instance.s8], [offset.s16]	*/
				/*	writes a 32 bit "single" to an instance						*/

    _ins_wdb,	/* write data-byte					[instance.s8], [offset.s16]	*/
				/*	writes a byte (8 bits) to an instance						*/

    _ins_wdfd,	/* write data-field decoded			Field0						*/
				/*	writes integer by a user-defined field (pulled offset/width)*/
   /* 0bxx011000 */
    _ins_rddu,	/* read data-double with update		[instance.s8], [offset.s16]	*/
				/*	moves instance by offset; reads double						*/

    _ins_rdsu,	/* read data-single with update		[instance.s8], [offset.s16]	*/
				/*	moves instance by offset; reads 32 bit "single"				*/

    _ins_rdbu,	/* read data-byte with update		[instance.s8], [offset.s16]	*/
				/*	moves instance by offset; reads byte						*/

    _ins_rdfe,	/* read data-field encoded			Field16						*/
				/*	reads integer data by an instruction-defined field			*/

    _ins_wddu,	/* write data-double with update	[instance.s8], [offset.s16]	*/
				/*	moves instance by offset; writes 64 bit "double"			*/

    _ins_wdsu,	/* write data-single with update	[instance.s8], [offset.s16]	*/
				/*	moves instance by offset; writes 32 bit "single"			*/

    _ins_wdbu,	/* write data-byte with update		[instance.s8], [offset.s16]	*/
				/*	moves instance by offset; writes byte (8 bits)				*/

    _ins_wdfe,	/* write data-field encoded			Field16						*/
				/*	writes integer data by an instruction-defined field			*/
   /* 0bxx100000 */
    _ins_or,	/* bitwise OR						(none)						*/
				/*	ds[1] |= ds[0]; ++ds; -- values are clipped to top 52 bits	*/

    _ins_and,	/* bitwise AND						(none)						*/
				/*	ds[1] &= ds[0]; ++ds; -- values are clipped to top 52 bits	*/

    _ins_eor,	/* bitwise Exclusive-OR				(none)						*/
				/*	ds[1] ^= ds[0]; ++ds; -- values are clipped to top 52 bits	*/

    _ins_sz,	/* set if zero						(none)						*/
				/*	ds[0] = !ds[0];												*/

    _ins_sgt,	/* set greater than					(none)						*/
				/*	ds[1] = ds[1]>ds[0]; ++ds;									*/

    _ins_sge,	/* set greater or equal				(none)						*/
				/*	ds[1] = ds[1]>=ds[0]; ++ds;									*/

    _ins_slt,	/* set less than					(none)						*/
				/*	ds[1] = ds[1]<ds[0]; ++ds;									*/

    _ins_sle,	/* set less or equal				(none)						*/
				/*	ds[1] = ds[1]<=ds[0]; ++ds;									*/
   /* 0bxx101000 */
    _ins_add,	/* add								(none)						*/
				/*	ds[1] += ds[0]; ++ds;										*/

    _ins_sub,	/* subtract							(none)						*/
				/*	ds[1] -= ds[0]; ++ds;										*/

    _ins_mul,	/* multiply							(none)						*/
				/*	ds[1] *= ds[0]; ++ds;										*/

    _ins_div,	/* divide							(none)						*/
				/*	ds[1] /= ds[0]; ++ds;										*/

    _ins_abs,	/* absolute value					(none)						*/
				/*	ds[0] = fabs(ds[0]);										*/

    _ins_scal,	/* scale by power of two			(none)						*/
				/*	ds[1] = scalb(ds[1],ds[0]); ++ds;							*/

    _ins_rndz,	/* round toward zero				(none)						*/
				/*	ds[0] = trunc(ds[0]);										*/

    _ins_sqrt,	/* square root						(none)						*/
				/*	sqrt(ds[0]);												*/
   /* 0bxx110000 */
    _ins_jn,	/* jump near						SInt24:offset				*/
				/*	jump to offset (always)										*/

    _ins_jnc,	/* jump near on condition			SInt24:offset				*/
				/*	jump to offset only if pulled data is non-zero				*/

    _ins_cn,	/* call near						SInt24:offset				*/
				/*	adds link, jumps to calculated address; clears the "result"	*/

    _ins_cnc,	/* call near on condition			SInt24:offset				*/
				/*	calls offset only if pulled data is non-zero				*/

    _ins_iad,	/* instance addr. dereference		[instance.s8], [offset.s16]	*/
				/*	pushes the address of an instance							*/

    _ins_iar,	/* instance addr. reference			[instance.s8], [offset.s16]	*/
				/*	sets an instance to the pulled absolute address				*/

    _ins_iaa,	/* instance address add				[instance.s8], [offset.s16]	*/
				/*	adds byte offset and pulled value to an instance			*/

    _ins_ci		/* call instance					[instance.s8], [offset.s16]	*/
				/*	calls instance address with (byte) offset					*/
   /* 0bxx111000 instructions are reserved for internal use */
};

enum
{ /* emulator errors */
	_err_null_pointer			= -1000,	/* parameter error */
	_err_floor_above_ceiling	= -1001,	/* parameter error */
	_err_bad_stack_position		= -1002,	/* runtime error (for protection) */
	_err_sensitive_register		= -1003		/* (for protection) */
};

/* ---------- structures */

typedef struct link_entry/*(16)*/
{
	SInt64			return_pc, return_ds;
} link_entry;

typedef struct sbcp_state/*(1084+)*/
{
	SInt64			instance[128];

	double			rs;

	/* the base pointer is the offset for all emulated pointers! */
	byte			*base_pointer;

	/* the cache map is a reference table to the cache tables (in the fake CPU) */
	/* each byte makes up for all mapped space in 65536 byte pages (16 bits) */

	/* for each representing byte: [writable.1] [table.7];
	  table 0bx0000000 is never mapped; all other values mean they "mapped" to a
	  cache table; 0bx1111111 refers to no table (uncached - but still mapped) */
	byte			*mr; /* cache map */
	UInt32			cache_page_count; /* number of representing bytes in map */

	/* current stack positions */
	double			*ds;
	link_entry		*ls;

	/* status register, program counter */
	UInt32			sr; /* [boundaries.1] [privilege.1] [reserved.3] [int_level.3] [zero.24] */
	byte			*pc; /* cannot be NULL or ==base_pointer */

	/* data/link floor/ceiling */
	double			*df, *dc;
	link_entry		*lf, *lc;

	/* other registers */
	UInt16			cs;
	UInt16			vs;
	SInt64			v1; /* points to the exception vector table */
} sbcp_state;

/* ---------- prototypes */

/* for maximum_lines, NULL represents "unlimited" */
/* the returned result is an error code or exception vector index (positive);
  whereas the exception couldn't be handled (and the emulator cannot advance) */
int sbcp_execute_lines(sbcp_state *state,
	long *maximum_lines, /* non-positive contents results in no execution */
	long int maximum_instructions); /* pass non-positive value to run indefinetly */
