/*
 sbcp_debug.c
  stack-based core prototype debugging section (optional)

 Created: January 26, 2007 3:09 AM EDST
 Revised: February 5, 2007 12:12 PM EDST

 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 *  This source code is offered for use in the public domain.  You may use,
 * modify or distribute it freely.
 *
 *  This code is distributed in the hope that it will be useful but WITHOUT
 * ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY DISCLAIMED.
 *  This includes but is not limited to warranties of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.

Friday, January 26, 2007 3:11:49 AM


[
  Current style:
	do
	{
		sbcp_disassemble(&position,buf,buf+64,buf+128,position-code);
		fprintf(stdout,"%s  %-19s|%s\n", buf,buf+64,buf+128);
	} while(--remaining!=0);
]
*/

#include <cseries.h>

#include <stdio.h>
#include <string.h>
#include <fp.h>

#include "sbcp_core.h"
#include "sbcp_lib.h"
#include "sbcp_debug.h"

/* ---------- globals */

/*
  each name <7 chars; the name covers a C-string; blanks are "undefined"

	for an opcode to have any operands here means that it is four bytes in
  size; otherwise it is always one byte

	each operand uses a char as a type:
		1 = 24 bit value (data)
		2 = 24 bit signed stack frame index
		3 = 24 bit signed offset
		4 = 24 bit Float24
		5 = instance record: [instance.s7] [offset.s16]
		6 = cache table record: [type.8] [offset.16]
		7 = register record: [which.8] [reserved.16]
		8 = field0: [sign.1] [instance.s7] [offset.s16]
		9 = field16: [sign.1] [instance.7] [offset.s8] [reserved.3], [width.5]
*/
typedef const char		opcode_info_data[8], register_name[8];

static opcode_info_data		opcodes[64] =
{
	"nop" "\0\1",
	"ads" "\0\1",
	"pshi" "\0\4",
	"nea" "\0\3",
	"stop",
	"ret",
	"pshz",
	"exg",

	"setrs" "\0\2",
	"getrs" "\0\2",
	"pshd" "\0\2",
	"popd" "\0\2",
	"orc" "\0\2",
	"andc" "\0\2",
	"cmpd" "\0\2",
	"addc" "\0\2",

	"rdd" "\0\5",
	"rds" "\0\5",
	"rdb" "\0\5",
	"rdfd" "\0\10",
	"wdd" "\0\5",
	"wds" "\0\5",
	"wdb" "\0\5",
	"wdfd" "\0\10",

	"rddu" "\0\5",
	"rdsu" "\0\5",
	"rdbu" "\0\5",
	"rdfe" "\0\11",
	"wddu" "\0\5",
	"wdsu" "\0\5",
	"wdbu" "\0\5",
	"wdfe" "\0\11",

	"or",
	"and",
	"eor",
	"sne",
	"sgt",
	"sge",
	"slt",
	"sle",

	"add",
	"sub",
	"mul",
	"div",
	"abs",
	"scal",
	"rndz",
	"sqrt",

	"jn" "\0\3",
	"jnc" "\0\3",
	"cn" "\0\3",
	"cnc" "\0\3",
	"iad" "\0\5",
	"iar" "\0\5",
	"iaa" "\0\5",
	"ci" "\0\5",

	"rct" "\0\6",
	"wct" "\0\6",
	"setrg" "\0\7",
	"getrg" "\0\7",
	"setds" "\0\1",
	"setpc" "\0\1",
	"",
	"illegal",
};

static register_name	register_names[16] =
{
	"DS", /* data stack register */
	"LS", /* link stack register */
	"SR", /* status register */
	"PC", /* program counter */

	"DF", /* data floor */
	"DC", /* data ceiling */
	"LF", /* link floor */
	"LC", /* link ceiling */

	"CR", /* cache register - always zero in emulation */
	"CS", /* cache status register - remains constant in emulation */
	"MR", /* map register */
	"VS", /* vector status register - remains constant in emulation */

	"V0", /* vector table 0 (execution vector) - always zero in emulation */
	"V1", /* vector table 1 (exception vector) */
	"V2", /* vector table 2 (currently undefined) */
	"V3", /* vector table 3 (currently undefined) */
};

/* ---------- code */

int sbcp_disassemble(byte **position,
	char *opname,
	char *operands,
	char *comment,
	long displacement)
{
	const char		*opcode_info;

	byte			*s, g;
	UInt32			opcode, t;

	if (position==(byte**)NULL || (s=*position)==(byte*)NULL)
		return -1;

	opcode = *s++;
	opcode_info = (const char*)((opcode&0x3F)+opcodes);

	if (operands!=(char*)NULL)
		operands[0] = 0;
	if (comment!=(char*)NULL)
		comment[0] = 0;
	if (opcode_info[0]==0)
		goto bad; /* this opcode is undefined */
	else
	{
		if (opname!=(char*)NULL)
			sprintf(opname,"%-8s%c%c", opcode_info,(opcode&0x80) ? 'B':' ',(opcode&0x40) ? 'L':' ');
		if ((t=strlen(opcode_info))<=6 && (t=opcode_info[t+1])!=0)
		{
			/* piece together the remaining bits of the opcode */
			g = opcode = *s++;
			opcode <<= 8; opcode |= *s++;
			opcode <<= 8; opcode |= *s++;

			switch(t)
			{
			case 1:
				if (operands!=(char*)NULL)
					sprintf(operands,"0x%06X", opcode);
				break;
			case 2:
				if (operands==(char*)NULL)
					break;
				if (opcode==0x00800000L)
					strcpy(operands,"(DS)");
				else if (!(g&0x80))
					sprintf(operands,"+%06X(DS)", opcode);
				else
					sprintf(operands,"-%06X(DS)", (-opcode)&0x00FFFFFFL);
				break;
			case 3:
				if (comment!=(char*)NULL)
					sprintf(comment," 0x%08X", ((signed)(opcode<<8)>>8)+displacement);
				if (!(g&0x80))
				{
					if (operands!=(char*)NULL)
						sprintf(operands,"*+%06X", opcode);
				} else if (operands!=(char*)NULL)
					sprintf(operands,"*-%06X", (-opcode)&0x00FFFFFFL);
				break;
			case 4:
				if (operands!=(char*)NULL)
					sprintf(operands,"#$\"%06X\"", opcode);
				if (comment==(char*)NULL)
					break;
				sprintf(comment,"%c%.7G", (g&0x80) ? '-':' ',
					scalb(0x00010000L+(opcode&0x0000FFFFL),(g&0x7F)-79));
				if (strchr(comment,'.')==(char*)NULL)
					strcat(comment,".0");
				break;
			case 5:
				if (g&0x80)
					goto bad;
				if (operands==(char*)NULL)
					break;
				if ((t=opcode&0x00008000L)!=0)
					opcode = (-opcode)&0x0000FFFFL;
				else
					opcode &= 0x0000FFFFL;
				sprintf(operands,"#0x%02X%c0x%04X", g,t ? '-':'+',opcode);
			case 6:
				if (operands==(char*)NULL)
					break;
				sprintf(operands,"#0x%02X+%04X", g,opcode&0x0000FFFFL);
				break;
			case 7:
				if ((opcode&0x00F0FFFFL)!=0)
					goto bad; /* currently an invalid opcode */
				strcpy(operands,register_names[g]);
				break;
			case 8:
				if (opcode&224)
					goto bad;
				if (operands==(char*)NULL)
					break;
				if (opcode&0x00008000L)
					t = (-opcode)&0x0000FFFFL;
				else
					t = opcode&0x0000FFFFL;
				sprintf(operands,"#0x%02X%c0x%04X{%c}", g,(opcode&0x00008000L) ?
					'-':'+',t,(opcode&0x00800000L) ? 'S':'U');
				break;
			case 9:
				if (opcode&0x008000E0L)
				{
				bad:
					if (opname!=(char*)NULL)
						strcpy(opname,"????");
					break;
				}
				if (operands==(char*)NULL)
					break;
				sprintf(operands,"#0x%02X%+d{%d:%c%d}", g,
					((signed)(opcode<<16)>>27),(opcode>>8)&3,(opcode&0x00800000L) ? 'S':'U',opcode&31);
				break;
			}
		}
	}

	*position = s;
	return 0;
}
